# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math
import itertools

import networkx as nx

from dwave_networkx.utils import binary_quadratic_model_sampler

__all__ = ["is_vertex_coloring",
           "is_cycle",
           "min_vertex_color",
           "min_vertex_coloring",  # alias for min_vertex_color
           "min_vertex_color_qubo",
           "vertex_color",
           "vertex_color_qubo",
           ]


@nx.utils.decorators.nodes_or_number(1)
def vertex_color_qubo(G, colors):
    """Return the QUBO with ground states corresponding to a vertex coloring.

    If `V` is the set of nodes, `E` is the set of edges and `C` is the set of
    colors the resulting qubo will have:

    * :math:`|V|*|C|` variables/nodes
    * :math:`|V|*|C|*(|C| - 1) / 2 + |E|*|C|` interactions/edges

    The QUBO has ground energy :math:`-|V|` and an infeasible gap of 1.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum vertex coloring.

    colors : int/sequence
        The colors. If an int, the colors are labelled `[0, n)`. The number of
        colors must be greater or equal to the chromatic number of the graph.

    Returns
    -------
    QUBO : dict
        The QUBO with ground states corresponding to valid colorings of the
        graph. The QUBO variables are labelled `(v, c)` where `v` is a node
        in `G` and `c` is a color. In the ground state of the QUBO, a variable
        `(v, c)` has value 1 if `v` should be colored `c` in a valid coloring.


    """
    _, colors = colors

    Q = {}

    # enforce that each variable in G has at most one color
    for v in G.nodes:
        # 1 in k constraint
        for c in colors:
            Q[(v, c), (v, c)] = -1

        for c0, c1 in itertools.combinations(colors, 2):
            Q[(v, c0), (v, c1)] = 2

    # enforce that adjacent nodes do not have the same color
    for u, v in G.edges:
        # NAND constraint
        for c in colors:
            Q[(u, c), (v, c)] = 1

    return Q


@binary_quadratic_model_sampler(2)
def vertex_color(G, colors, sampler=None, **sampler_args):
    """Returns an approximate vertex coloring.

    Vertex coloring is the problem of assigning a color to the
    vertices of a graph in a way that no adjacent vertices have the
    same color.

    Defines a QUBO [DWMP]_ with ground states corresponding to valid
    vertex colorings and uses the sampler to sample from it.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum vertex coloring.

    colors : int/sequence
        The colors. If an int, the colors are labelled `[0, n)`. The number of
        colors must be greater or equal to the chromatic number of the graph.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    coloring : dict
        A coloring for each vertex in G such that no adjacent nodes
        share the same color. A dict of the form {node: color, ...}


    References
    ----------
    .. [DWMP] Dahl, E., "Programming the D-Wave: Map Coloring Problem",
       https://www.dwavesys.com/media/htfgw5bk/map-coloring-wp2.pdf

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """
    Q = vertex_color_qubo(G, colors)

    # get the lowest energy sample
    sample = sampler.sample_qubo(Q, **sampler_args).first.sample

    return {v: c for (v, c), val in sample.items() if val}


def _chromatic_number_upper_bound(G):
    # tries to determine an upper bound on the chromatic number of G
    # Assumes G is not complete

    if not nx.is_connected(G):
        return max((_chromatic_number_upper_bound(G.subgraph(c))
                    for c in nx.connected_components(G)))

    n_nodes = len(G.nodes)
    n_edges = len(G.edges)

    # chi * (chi - 1) <= 2 * |E|
    quad_bound = math.ceil((1 + math.sqrt(1 + 8 * n_edges)) / 2)

    if n_nodes % 2 == 1 and is_cycle(G):
        # odd cycle graphs need three colors
        bound = 3
    elif n_nodes > 2:
        try:
            import numpy as np
        except ImportError:
            # chi <= max degree, unless it is complete or a cycle graph of odd length,
            # in which case chi <= max degree + 1 (Brook's Theorem)
            bound = max(G.degree(node) for node in G)
        else:
            # Let A be the adj matrix of G (symmetric, 0 on diag). Let theta_1
            # be the largest eigenvalue of A. Then chi <= theta_1 + 1 with
            # equality iff G is complete or an odd cycle.
            # this is strictly better than brooks theorem
            bound = math.ceil(max(np.linalg.eigvals(nx.to_numpy_array(G))))
    else:
        # we know it's connected
        bound = n_nodes

    return min(quad_bound, bound)


def _chromatic_number_lower_bound(G):
    # find a random maximal clique and use that to determine a lower bound
    v = max(G, key=G.degree)

    clique = {v}
    for u in G[v]:
        if all(w in G[u] for w in clique):
            clique.add(u)

    return len(clique)


def min_vertex_color_qubo(G, chromatic_lb=None, chromatic_ub=None):
    """Return a QUBO with ground states corresponding to a minimum vertex
    coloring.

    Vertex coloring is the problem of assigning a color to the
    vertices of a graph in a way that no adjacent vertices have the
    same color. A minimum vertex coloring is the problem of solving
    the vertex coloring problem using the smallest number of colors.

    Defines a QUBO [DWMP]_ with ground states corresponding to minimum
    vertex colorings and uses the sampler to sample from it.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum vertex coloring.

    chromatic_lb : int, optional
         A lower bound on the chromatic number. If one is not provided, a
         bound is calulcated.

    chromatic_ub : int, optional
        An upper bound on the chromatic number. If one is not provided, a bound
        is calculated.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    QUBO : dict
        The QUBO with ground states corresponding to minimum colorings of the
        graph. The QUBO variables are labelled `(v, c)` where `v` is a node
        in `G` and `c` is a color. In the ground state of the QUBO, a variable
        `(v, c)` has value 1 if `v` should be colored `c` in a valid coloring.

    """

    chi_ub = _chromatic_number_upper_bound(G)
    chromatic_ub = chi_ub if chromatic_ub is None else min(chi_ub, chromatic_ub)

    chib_lb = _chromatic_number_lower_bound(G)
    chromatic_lb = chib_lb if chromatic_lb is None else max(chib_lb, chromatic_lb)

    if chromatic_lb > chromatic_ub:
        raise RuntimeError("something went wrong when calculating the "
                           "chromatic number bounds")

    # our base QUBO is one with as many colors as we might need, so we use the
    # upper bound
    Q = vertex_color_qubo(G, int(chromatic_ub))

    if chromatic_lb != chromatic_ub:
        # we want to penalize the colors that we aren't sure that we need
        # we might need to use some of the colors, so we want to penalize
        # them in increasing amounts, linearly.

        num_penalized = chromatic_ub - chromatic_lb

        # we want evenly spaced penalties in (0, 1) without the endpoints
        weights = [p / (num_penalized + 1) for p in range(1, num_penalized + 1)]

        for p, c in zip(weights, range(chromatic_lb, chromatic_ub)):
            for v in G:
                Q[(v, c), (v, c)] += p

    return Q


@binary_quadratic_model_sampler(1)
def min_vertex_color(G, sampler=None, chromatic_lb=None, chromatic_ub=None,
                     **sampler_args):
    """Returns an approximate minimum vertex coloring.

    Vertex coloring is the problem of assigning a color to the
    vertices of a graph in a way that no adjacent vertices have the
    same color. A minimum vertex coloring is the problem of solving
    the vertex coloring problem using the smallest number of colors.

    Defines a QUBO [DWMP]_ with ground states corresponding to minimum
    vertex colorings and uses the sampler to sample from it.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which to find a minimum vertex coloring.

    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy. If no
        sampler is provided, one must be provided using the
        `set_default_sampler` function.

    chromatic_lb : int, optional
         A lower bound on the chromatic number. If one is not provided, a
         bound is calulcated.

    chromatic_ub : int, optional
        An upper bound on the chromatic number. If one is not provided, a bound
        is calculated.

    sampler_args
        Additional keyword parameters are passed to the sampler.

    Returns
    -------
    coloring : dict
        A coloring for each vertex in G such that no adjacent nodes
        share the same color. A dict of the form {node: color, ...}

    References
    ----------
    .. [DWMP] Dahl, E., "Programming the D-Wave: Map Coloring Problem",
       https://www.dwavesys.com/media/htfgw5bk/map-coloring-wp2.pdf

    Notes
    -----
    Samplers by their nature may not return the optimal solution. This
    function does not attempt to confirm the quality of the returned
    sample.

    """

    Q = min_vertex_color_qubo(G, chromatic_lb=chromatic_lb,
                              chromatic_ub=chromatic_ub)

    # get the lowest energy sample
    sample = sampler.sample_qubo(Q, **sampler_args).first.sample

    return {v: c for (v, c), val in sample.items() if val}


# legacy name, alias
min_vertex_coloring = min_vertex_color


def is_cycle(G):
    """Determines whether the given graph is a cycle or circle graph.

    A cycle graph or circular graph is a graph that consists of a single cycle.

    https://en.wikipedia.org/wiki/Cycle_graph

    Parameters
    ----------
    G : NetworkX graph

    Returns
    -------
    is_cycle : bool
        True if the graph consists of a single cycle.

    """
    if len(G) <= 2:
        return False

    trailing, leading = next(iter(G.edges))
    start_node = trailing

    # travel around the graph, checking that each node has degree exactly two
    # also track how many nodes were visited
    n_visited = 1
    while leading != start_node:
        neighbors = G[leading]

        if len(neighbors) != 2:
            return False

        node1, node2 = neighbors

        if node1 == trailing:
            trailing, leading = leading, node2
        else:
            trailing, leading = leading, node1

        n_visited += 1

    # if we haven't visited all of the nodes, then it is not a connected cycle
    return n_visited == len(G)


def is_vertex_coloring(G, coloring):
    """Determines whether the given coloring is a vertex coloring of graph G.

    Parameters
    ----------
    G : NetworkX graph
        The graph on which the vertex coloring is applied.

    coloring : dict
        A coloring of the nodes of G. Should be a dict of the form
        {node: color, ...}.

    Returns
    -------
    is_vertex_coloring : bool
        True if the given coloring defines a vertex coloring; that is, no
        two adjacent vertices share a color.

    Example
    -------
    This example colors checks two colorings for a graph, G, of a single Chimera
    unit cell. The first uses one color (0) for the four horizontal qubits
    and another (1) for the four vertical qubits, in which case there are
    no adjacencies; the second coloring swaps the color of one node.

    >>> G = dnx.chimera_graph(1,1,4)
    >>> colors = {0: 0, 1: 0, 2: 0, 3: 0, 4: 1, 5: 1, 6: 1, 7: 1}
    >>> dnx.is_vertex_coloring(G, colors)
    True
    >>> colors[4]=0
    >>> dnx.is_vertex_coloring(G, colors)
    False

   """
    return all(coloring[u] != coloring[v] for u, v in G.edges)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import dimod

import dwave_networkx as dnx


class Test_sample_markov_network_bqm(unittest.TestCase):
    def test_one_node(self):
        potentials = {'a': {(0,): 1.2, (1,): .4}}

        bqm = dnx.markov_network_bqm(dnx.markov_network(potentials))

        for edge, potential in potentials.items():
            for config, energy in potential.items():
                sample = dict(zip(edge, config))
                self.assertAlmostEqual(bqm.energy(sample), energy)

    def test_one_edge(self):
        potentials = {'ab': {(0, 0): 1.2, (1, 0): .4,
                             (0, 1): 1.3, (1, 1): -4}}

        bqm = dnx.markov_network_bqm(dnx.markov_network(potentials))

        for edge, potential in potentials.items():
            for config, energy in potential.items():
                sample = dict(zip(edge, config))
                self.assertAlmostEqual(bqm.energy(sample), energy)

    def test_typical(self):
        potentials = {'a': {(0,): 1.5, (1,): -.5},
                      'ab': {(0, 0): 1.2, (1, 0): .4,
                             (0, 1): 1.3, (1, 1): -4},
                      'bc': {(0, 0): 1.7, (1, 0): .4,
                             (0, 1): -1, (1, 1): -4},
                      'd': {(0,): -.5, (1,): 1.6}}

        bqm = dnx.markov_network_bqm(dnx.markov_network(potentials))

        samples = dimod.ExactSolver().sample(bqm)

        for sample, energy in samples.data(['sample', 'energy']):

            en = 0
            for interaction, potential in potentials.items():
                config = tuple(sample[v] for v in interaction)
                en += potential[config]

            self.assertAlmostEqual(en, energy)


class Test_sample_markov_network(unittest.TestCase):
    def test_typical_return_sampleset(self):
        potentials = {'a': {(0,): 1.5, (1,): -.5},
                      'ab': {(0, 0): 1.2, (1, 0): .4,
                             (0, 1): 1.3, (1, 1): -4},
                      'bc': {(0, 0): 1.7, (1, 0): .4,
                             (0, 1): -1, (1, 1): -4},
                      'd': {(0,): -.5, (1,): 1.6}}

        MN = dnx.markov_network(potentials)

        samples = dnx.sample_markov_network(MN, dimod.ExactSolver(),
                                            fixed_variables={'c': 0},
                                            return_sampleset=True)

        for sample, energy in samples.data(['sample', 'energy']):
            self.assertEqual(sample['c'], 0)

            en = 0
            for interaction, potential in potentials.items():
                config = tuple(sample[v] for v in interaction)
                en += potential[config]

            self.assertAlmostEqual(en, energy)

    def test_typical(self):
        potentials = {'a': {(0,): 1.5, (1,): -.5},
                      'ab': {(0, 0): 1.2, (1, 0): .4,
                             (0, 1): 1.3, (1, 1): -4},
                      'bc': {(0, 0): 1.7, (1, 0): .4,
                             (0, 1): -1, (1, 1): -4},
                      'd': {(0,): -.5, (1,): 1.6}}

        MN = dnx.markov_network(potentials)

        bqm = dnx.markov_network_bqm(MN)

        samples = dnx.sample_markov_network(MN, dimod.ExactSolver(),
                                            fixed_variables={'c': 0})

        for sample in samples:
            self.assertEqual(sample['c'], 0)

            en = 0
            for interaction, potential in potentials.items():
                config = tuple(sample[v] for v in interaction)
                en += potential[config]

            self.assertAlmostEqual(en, bqm.energy(sample))



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Sets a binary quadratic model sampler used by default
for functions that require a sample when none is specified.

A sampler is a process that samples
from low-energy states in models defined by an Ising equation
or a Quadratic Unconstrained Binary Optimization Problem
(QUBO).

Sampler API
-----------
* Required Methods: 'sample_qubo' and 'sample_ising'
* Return value: iterable of samples, in order of increasing energy

See  `dimod <https://github.com/dwavesystems/dimod>`_ for details.

Example
-------
This example creates and uses a placeholder for binary quadratic model
samplers that returns a correct response only in the case of finding an
independent set on a complete graph (where one node is always an
independent set). The placeholder sampler can be used to test the simple
examples of the functions for configuring a default sampler.

>>> # Create a placeholder sampler
>>> class ExampleSampler:
...     # an example sampler, only works for independent set on complete
...     # graphs
...     def __init__(self, name):
...         self.name = name
...     def sample_ising(self, h, J):
...         sample = {v: -1 for v in h}
...         sample[0] = 1  # set one node to true
...         return [sample]
...     def sample_qubo(self, Q):
...         sample = {v: 0 for v in set().union(*Q)}
...         sample[0] = 1  # set one node to true
...         return [sample]
...     def __str__(self):
...         return self.name
...
>>> # Identify the new sampler as the default sampler
>>> sampler0 = ExampleSampler('sampler0')
>>> dnx.set_default_sampler(sampler0)
>>> # Find an independent set using the default sampler
>>> G = nx.complete_graph(5)
>>> dnx.maximum_independent_set(G)
[0]

"""

from dwave_networkx.utils.decorators import binary_quadratic_model_sampler

__all__ = ['set_default_sampler', 'get_default_sampler', 'unset_default_sampler']


_SAMPLER = None


@binary_quadratic_model_sampler(0)
def set_default_sampler(sampler):
    """Sets a default binary quadratic model sampler.

    Parameters
    ----------
    sampler
        A binary quadratic model sampler. A sampler is a process that
        samples from low-energy states in models defined by an Ising
        equation or a Quadratic Unconstrained Binary Optimization
        Problem (QUBO). A sampler is expected to have a 'sample_qubo'
        and 'sample_ising' method. A sampler is expected to return an
        iterable of samples, in order of increasing energy.

    Examples
    --------
    This example sets sampler0 as the default sampler and finds an independent
    set for graph G, first using the default sampler and then overriding it by
    specifying a different sampler.

    >>> dnx.set_default_sampler(sampler0)  # doctest: +SKIP
    >>> indep_set = dnx.maximum_independent_set_dm(G)  # doctest: +SKIP
    >>> indep_set = dnx.maximum_independent_set_dm(G, sampler1)  # doctest: +SKIP

    """
    global _SAMPLER
    _SAMPLER = sampler


def unset_default_sampler():
    """Resets the default sampler back to None.

    Examples
    --------
    This example sets sampler0 as the default sampler, verifies the setting,
    then resets the default, and verifies the resetting.

    >>> dnx.set_default_sampler(sampler0)  # doctest: +SKIP
    >>> print(dnx.get_default_sampler())  # doctest: +SKIP
    'sampler0'
    >>> dnx.unset_default_sampler()  # doctest: +SKIP
    >>> print(dnx.get_default_sampler())  # doctest: +SKIP
    None
    """
    global _SAMPLER
    _SAMPLER = None


def get_default_sampler():
    """Queries the current default sampler.

    Examples
    --------
    This example queries the default sampler before and after specifying
    a default sampler.

    >>> print(dnx.get_default_sampler())  # doctest: +SKIP
    None
    >>> dnx.set_default_sampler(sampler)  # doctest: +SKIP
    >>> print(dnx.get_default_sampler())  # doctest: +SKIP
    'sampler'

    """
    return _SAMPLER



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dwave_networkx as dnx

alpha_map = dict(enumerate('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'))


class TestChimeraGraph(unittest.TestCase):
    def test_single_tile(self):

        # fully specified
        G = dnx.chimera_graph(1, 1, 4)

        # should have 8 nodes
        self.assertEqual(len(G), 8)

        # nodes 0,...,7 should be in the tile
        for n in range(8):
            self.assertIn(n, G)

        # check bipartite
        for i in range(4):
            for j in range(4, 8):
                self.assertTrue((i, j) in G.edges() or (j, i) in G.edges())

    def test_find_chimera_indices_single_tile(self):
        for k in range(1, 10):
            G = dnx.chimera_graph(1, 1, k)

            # get the chimera indices, check that they match the ones already
            # present
            chimera_indices = dnx.find_chimera_indices(G)
            self._check_matching_chimera_indices(G, chimera_indices)

    def test_find_chimera_indices_single_tile_alpha_labels(self):
        for k in range(1, 10):
            G = nx.relabel_nodes(dnx.chimera_graph(1, 1, k), alpha_map)

            # get the chimera indices, check that they match the ones already
            # present
            chimera_indices = dnx.find_chimera_indices(G)
            self._check_matching_chimera_indices(G, chimera_indices)

    def test_find_chimera_indices_one_tile_degenerate(self):
        G = dnx.chimera_graph(1, 1, 5)

        # remove 1 node
        G.remove_node(4)
        chimera_indices = dnx.find_chimera_indices(G)
        self._check_matching_chimera_indices(G, chimera_indices)

        # remove another node
        G.remove_node(3)
        chimera_indices = dnx.find_chimera_indices(G)
        self._check_matching_chimera_indices(G, chimera_indices)

    def test_not_full_yield(self):
        edges = [(0, 3), (0, 2), (1, 3)]
        G = dnx.chimera_graph(1, 1, 2, edge_list=edges)
        for e in edges:
            self.assertIn(e, G.edges())
        for (u, v) in G.edges:
            self.assertTrue((u, v) in edges or (v, u) in edges)

        nodes = [0, 1, 2]
        G = dnx.chimera_graph(1, 1, 2, node_list=nodes)
        self.assertEqual(len(G), 3)
        self.assertEqual(len(G.edges()), 2)

        edges = [(0, 2), (1, 2)]
        nodes = [0, 1, 2, 3]
        G = dnx.chimera_graph(1, 1, 2, node_list=nodes, edge_list=edges)
        # 3 should be added as a singleton
        self.assertEqual(len(G[3]), 0)

    def test_float_robustness(self):
        G = dnx.chimera_graph(8 / 2)

        self.assertEqual(set(G.nodes), set(dnx.chimera_graph(4).nodes))
        for u, v in dnx.chimera_graph(4).edges:
            self.assertIn(u, G[v])

        G = dnx.chimera_graph(4, 4.)

        self.assertEqual(set(G.nodes), set(dnx.chimera_graph(4).nodes))
        for u, v in dnx.chimera_graph(4).edges:
            self.assertIn(u, G[v])

        G = dnx.chimera_graph(4, 4, 4.)

        self.assertEqual(set(G.nodes), set(dnx.chimera_graph(4).nodes))
        for u, v in dnx.chimera_graph(4).edges:
            self.assertIn(u, G[v])

    # def test_find_chimera_indices_typical(self):
    #     for t in range(2, 5):
    #         G = dnx.chimera_graph(2, 2, t)
    #         chimera_indices = dnx.find_chimera_indices(G)
    #         self._check_matching_chimera_indices(G, chimera_indices)

    #         G = dnx.chimera_graph(4, 4, t)
    #         chimera_indices = dnx.find_chimera_indices(G)
    #         self._check_matching_chimera_indices(G, chimera_indices)

    # def test_find_chimera_indices_shore_1(self):
    #     G = dnx.chimera_graph(2, 2, 1)
    #     chimera_indices = dnx.find_chimera_indices(G)
    #     self._check_matching_chimera_indices(G, chimera_indices)

    #     G = dnx.chimera_graph(4, 4, 1)
    #     chimera_indices = dnx.find_chimera_indices(G)
    #     self._check_matching_chimera_indices(G, chimera_indices)

    def _check_matching_chimera_indices(self, G, chimera_indices):
        for v, dat in G.nodes(data=True):
            self.assertEqual(dat['chimera_index'], chimera_indices[v])

    def test_coordinate_basics(self):
        from dwave_networkx.generators.chimera import chimera_coordinates
        G = dnx.chimera_graph(4)
        H = dnx.chimera_graph(4, coordinates=True)
        coords = chimera_coordinates(4)
        Gnodes = G.nodes
        Hnodes = H.nodes
        for v in Gnodes:
            q = Gnodes[v]['chimera_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.chimera_to_linear(q))
            self.assertEqual(q, coords.linear_to_chimera(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['chimera_index'], q)
            self.assertEqual(v, coords.chimera_to_linear(q))
            self.assertEqual(q, coords.linear_to_chimera(v))

    def test_coordinate_subgraphs(self):
        from dwave_networkx.generators.chimera import chimera_coordinates
        from random import sample
        G = dnx.chimera_graph(4)
        H = dnx.chimera_graph(4, coordinates=True)
        coords = chimera_coordinates(4)

        lmask = sample(list(G.nodes()), G.number_of_nodes()//2)
        cmask = list(coords.iter_linear_to_chimera(lmask))

        self.assertEqual(lmask, list(coords.iter_chimera_to_linear(cmask)))

        Gm = dnx.chimera_graph(4, node_list=lmask)
        Hm = dnx.chimera_graph(4, node_list=cmask, coordinates=True)

        Gs = G.subgraph(lmask)
        Hs = H.subgraph(cmask)

        EG = sorted(map(sorted, Gs.edges()))
        EH = sorted(map(sorted, Hs.edges()))

        self.assertEqual(EG, sorted(map(sorted, Gm.edges())))
        self.assertEqual(EH, sorted(map(sorted, Hm.edges())))

        Gn = dnx.chimera_graph(4, edge_list=EG)
        Hn = dnx.chimera_graph(4, edge_list=EH, coordinates=True)

        Gnodes = Gn.nodes
        Hnodes = Hn.nodes
        for v in Gnodes:
            q = Gnodes[v]['chimera_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.chimera_to_linear(q))
            self.assertEqual(q, coords.linear_to_chimera(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['chimera_index'], q)
            self.assertEqual(v, coords.chimera_to_linear(q))
            self.assertEqual(q, coords.linear_to_chimera(v))

        self.assertEqual(EG, sorted(map(sorted, coords.iter_chimera_to_linear_pairs(Hn.edges()))))
        self.assertEqual(EH, sorted(map(sorted, coords.iter_linear_to_chimera_pairs(Gn.edges()))))

    def test_linear_to_chimera(self):
        G = dnx.linear_to_chimera(212, 8, 8, 4)
        self.assertEqual(G, (3, 2, 1, 0))

    def test_chimera_to_linear(self):
        G = dnx.chimera_to_linear(3, 2, 1, 0, 8, 8, 4)
        self.assertEqual(G, 212)

    def test_nonsquare_coordinate_generator(self):
        #issue 149 found an issue with non-square generators -- let's be extra careful here
        for (m, n) in [(2, 4), (4, 2)]:
            G = dnx.chimera_graph(m, n, coordinates=True, data=True)
            H = dnx.chimera_graph(m, n, coordinates=False, data=True)
            self.assertTrue(nx.is_isomorphic(G, H))

            Gnodes = set(G.nodes)
            Glabels = set(q['linear_index'] for q in G.nodes.values())

            Hnodes = set(H.nodes)
            Hlabels = set(q['chimera_index'] for q in H.nodes.values())

            self.assertEqual(Gnodes, Hlabels)
            self.assertEqual(Hnodes, Glabels)

            coords = dnx.chimera_coordinates(m, n)
            F = nx.relabel_nodes(G, coords.chimera_to_linear, copy=True)
            self.assertEqual(set(map(frozenset, F.edges)), set(map(frozenset, H.edges)))

            E = nx.relabel_nodes(H, coords.linear_to_chimera, copy=True)
            self.assertEqual(set(map(frozenset, E.edges)), set(map(frozenset, G.edges)))


    def test_graph_relabeling(self):
        def graph_equal(g, h):
            self.assertEqual(set(g), set(h))
            self.assertEqual(
                set(map(tuple, map(sorted, g.edges))),
                set(map(tuple, map(sorted, g.edges)))
            )
            for v, d in g.nodes(data=True):
                self.assertEqual(h.nodes[v], d)

        coords = dnx.chimera_coordinates(3)
        for data in True, False:
            c3l = dnx.chimera_graph(3, data=data)
            c3c = dnx.chimera_graph(3, data=data, coordinates=True)

            graph_equal(c3l, coords.graph_to_linear(c3c))
            graph_equal(c3l, coords.graph_to_linear(c3l))
            graph_equal(c3c, coords.graph_to_chimera(c3l))
            graph_equal(c3c, coords.graph_to_chimera(c3c))

        h = dnx.chimera_graph(2)
        del h.graph['labels']
        with self.assertRaises(ValueError):
            coords.graph_to_linear(h)
        with self.assertRaises(ValueError):
            coords.graph_to_chimera(h)


    def test_sublattice_mappings(self):
        def check_subgraph_mapping(f, g, h):
            for v in g:
                if not h.has_node(f(v)):
                    raise RuntimeError(f"node {v} mapped to {f(v)} is not in {h.graph['name']} ({h.graph['labels']})")
            for u, v in g.edges:
                if not h.has_edge(f(u), f(v)):
                    raise RuntimeError(f"edge {(u, v)} mapped to {(f(u), f(v))} not present in {h.graph['name']} ({h.graph['labels']})")

        c2l = dnx.chimera_graph(2)
        c2c = dnx.chimera_graph(2, coordinates=True)
        c32l = dnx.chimera_graph(3, 2)
        c23c = dnx.chimera_graph(2, 3, coordinates=True)

        c5l = dnx.chimera_graph(5)
        c5c = dnx.chimera_graph(5, coordinates=True)
        c54l = dnx.chimera_graph(5, 4)
        c45c = dnx.chimera_graph(4, 5, coordinates=True)

        for target in c5l, c5c, c54l, c45c:
            for source in c2l, c2c, c32l, c23c, target:
                covered = set()
                for f in dnx.chimera_sublattice_mappings(source, target):
                    check_subgraph_mapping(f, source, target)
                    covered.update(map(f, source))
                self.assertEqual(covered, set(target))




# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import warnings

from random import sample

import networkx as nx
import dwave_networkx as dnx

from dwave_networkx.generators.pegasus import (
    fragmented_edges,
    get_tuple_defragmentation_fn,
    get_tuple_fragmentation_fn,
    )

alpha_map = dict(enumerate('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'))


class TestPegasusGraph(unittest.TestCase):
    def test_p2(self):
        G = dnx.pegasus_graph(2, fabric_only=False)

        # should have 48 nodes
        self.assertEqual(len(G), 48)

        # nodes 0,...,47 should be in the graph
        for n in range(48):
            self.assertIn(n, G)

    def test_bad_args(self):
        with self.assertRaises(dnx.DWaveNetworkXException):
            G = dnx.pegasus_graph(2, offset_lists=[], offsets_index=0)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            G = dnx.pegasus_graph(2, offset_lists=[[0, 1]*6, [0, 1]*6])
            self.assertLessEqual(len(w), 13)
            self.assertGreaterEqual(len(w), 12)


class TestPegasusCoordinates(unittest.TestCase):

    def test_connected_component(self):
        test_offsets = [[0] * 12] * 2, [[2] * 12, [6] * 12], [[6] * 12, [2, 2, 6, 6, 10, 10] * 2], [[2, 2, 6, 6, 10, 10] * 2] * 2
        for offsets in test_offsets:
            G = dnx.pegasus_graph(4, fabric_only=True, offset_lists=offsets)
            H = dnx.pegasus_graph(4, fabric_only=False, offset_lists=offsets)
            nodes = sorted(G)
            comp = sorted(max((G.subgraph(c).copy() for c in nx.connected_components(G)), key=len))
            self.assertEqual(comp, nodes)

    def test_coordinate_basics(self):
        G = dnx.pegasus_graph(4, fabric_only=False)
        H = dnx.pegasus_graph(4, coordinates=True, fabric_only=False)
        coords = dnx.pegasus_coordinates(4)
        Gnodes = G.nodes
        Hnodes = H.nodes
        for v in Gnodes:
            q = Gnodes[v]['pegasus_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.pegasus_to_linear(q))
            self.assertEqual(q, coords.linear_to_pegasus(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['pegasus_index'], q)
            self.assertEqual(v, coords.pegasus_to_linear(q))
            self.assertEqual(q, coords.linear_to_pegasus(v))

    def test_nice_coordinates(self):
        G = dnx.pegasus_graph(4, nice_coordinates=True)
        H = dnx.chimera_graph(3, coordinates=True)
        for p, q in H.edges():
            for t in range(3):
                pg = (t,) + p
                qg = (t,) + q
                self.assertTrue(G.has_edge(pg, qg))
        coords = dnx.pegasus_coordinates(4)
        n2p = coords.nice_to_pegasus
        p2n = coords.pegasus_to_nice
        n2l = coords.nice_to_linear
        l2n = coords.linear_to_nice
        for p in G.nodes():
            self.assertEqual(p2n(n2p(p)), p)
            self.assertEqual(l2n(n2l(p)), p)
            self.assertTrue(H.has_node(p[1:]))

        G = dnx.pegasus_graph(4)
        for p in G.nodes():
            self.assertEqual(n2l(l2n(p)), p)

        G = dnx.pegasus_graph(4, coordinates=True)
        for p in G.nodes():
            self.assertEqual(n2p(p2n(p)), p)


    def test_consistent_linear_nice_pegasus(self):
        P4 = dnx.pegasus_graph(4, nice_coordinates=True)

        coords = dnx.pegasus_coordinates(4)

        # `.*_to_*` methods

        for n, data in P4.nodes(data=True):
            p = data['pegasus_index']
            i = data['linear_index']
            self.assertEqual(coords.nice_to_pegasus(n), p)
            self.assertEqual(coords.pegasus_to_nice(p), n)
            self.assertEqual(coords.nice_to_linear(n), i)
            self.assertEqual(coords.linear_to_nice(i), n)
            self.assertEqual(coords.linear_to_pegasus(i), p)
            self.assertEqual(coords.pegasus_to_linear(p), i)

        # `.iter_*_to_*` methods

        nlist, plist, ilist = zip(*((n, d['pegasus_index'], d['linear_index'])
                                    for n, d in P4.nodes(data=True)))
        self.assertEqual(tuple(coords.iter_nice_to_pegasus(nlist)), plist)
        self.assertEqual(tuple(coords.iter_pegasus_to_nice(plist)), nlist)
        self.assertEqual(tuple(coords.iter_nice_to_linear(nlist)), ilist)
        self.assertEqual(tuple(coords.iter_linear_to_nice(ilist)), nlist)
        self.assertEqual(tuple(coords.iter_pegasus_to_linear(plist)), ilist)
        self.assertEqual(tuple(coords.iter_linear_to_pegasus(ilist)), plist)

        # `.iter_*_to_*_pairs` methods

        nedgelist = []
        pedgelist = []
        iedgelist = []
        for u, v in P4.edges:
            nedgelist.append((u, v))
            pedgelist.append((P4.nodes[u]['pegasus_index'],
                              P4.nodes[v]['pegasus_index']))
            iedgelist.append((P4.nodes[u]['linear_index'],
                              P4.nodes[v]['linear_index']))

        self.assertEqual(list(coords.iter_nice_to_pegasus_pairs(nedgelist)),
                         pedgelist)
        self.assertEqual(list(coords.iter_pegasus_to_nice_pairs(pedgelist)),
                         nedgelist)
        self.assertEqual(list(coords.iter_nice_to_linear_pairs(nedgelist)),
                         iedgelist)
        self.assertEqual(list(coords.iter_linear_to_nice_pairs(iedgelist)),
                         nedgelist)
        self.assertEqual(list(coords.iter_pegasus_to_linear_pairs(pedgelist)),
                         iedgelist)
        self.assertEqual(list(coords.iter_linear_to_pegasus_pairs(iedgelist)),
                         pedgelist)

    def test_coordinate_subgraphs(self):
        G = dnx.pegasus_graph(4)
        H = dnx.pegasus_graph(4, coordinates=True)
        coords = dnx.pegasus_coordinates(4)

        lmask = sample(list(G.nodes()), G.number_of_nodes()//2)
        cmask = list(coords.iter_linear_to_pegasus(lmask))

        self.assertEqual(lmask, list(coords.iter_pegasus_to_linear(cmask)))

        Gm = dnx.pegasus_graph(4, node_list=lmask)
        Hm = dnx.pegasus_graph(4, node_list=cmask, coordinates=True)

        Gs = G.subgraph(lmask)
        Hs = H.subgraph(cmask)

        EG = sorted(map(sorted, Gs.edges()))
        EH = sorted(map(sorted, Hs.edges()))

        self.assertEqual(EG, sorted(map(sorted, Gm.edges())))
        self.assertEqual(EH, sorted(map(sorted, Hm.edges())))

        Gn = dnx.pegasus_graph(4, edge_list=EG)
        Hn = dnx.pegasus_graph(4, edge_list=EH, coordinates=True)

        Gnodes = Gn.nodes
        Hnodes = Hn.nodes
        for v in Gnodes:
            q = Gnodes[v]['pegasus_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.pegasus_to_linear(q))
            self.assertEqual(q, coords.linear_to_pegasus(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['pegasus_index'], q)
            self.assertEqual(v, coords.pegasus_to_linear(q))
            self.assertEqual(q, coords.linear_to_pegasus(v))

        self.assertEqual(EG, sorted(map(sorted, coords.iter_pegasus_to_linear_pairs(Hn.edges()))))
        self.assertEqual(EH, sorted(map(sorted, coords.iter_linear_to_pegasus_pairs(Gn.edges()))))

    def test_graph_relabeling(self):
        def graph_equal(g, h):
            self.assertEqual(set(g), set(h))
            self.assertEqual(
                set(map(tuple, map(sorted, g.edges))),
                set(map(tuple, map(sorted, g.edges)))
            )
            for v, d in g.nodes(data=True):
                self.assertEqual(h.nodes[v], d)

        coords = dnx.pegasus_coordinates(3)
        nodes_nice = dnx.pegasus_graph(3, nice_coordinates=True)
        nodes_linear = list(coords.iter_nice_to_linear(nodes_nice))
        nodes_pegasus = list(coords.iter_nice_to_pegasus(nodes_nice))
        
        for data in True, False:
            p3l = dnx.pegasus_graph(3, data=data).subgraph(nodes_linear)
            p3p = dnx.pegasus_graph(3, data=data, coordinates=True).subgraph(nodes_pegasus)
            p3n = dnx.pegasus_graph(3, data=data, nice_coordinates=True)

            graph_equal(p3l, coords.graph_to_linear(p3l))
            graph_equal(p3l, coords.graph_to_linear(p3p))
            graph_equal(p3l, coords.graph_to_linear(p3n))
            
            graph_equal(p3p, coords.graph_to_pegasus(p3l))
            graph_equal(p3p, coords.graph_to_pegasus(p3p))
            graph_equal(p3p, coords.graph_to_pegasus(p3n))

            graph_equal(p3n, coords.graph_to_nice(p3l))
            graph_equal(p3n, coords.graph_to_nice(p3p))
            graph_equal(p3n, coords.graph_to_nice(p3n))

        h = dnx.pegasus_graph(2)
        del h.graph['labels']
        with self.assertRaises(ValueError):
            coords.graph_to_nice(h)
        with self.assertRaises(ValueError):
            coords.graph_to_linear(h)
        with self.assertRaises(ValueError):
            coords.graph_to_pegasus(h)

    def test_sublattice_mappings(self):
        def check_subgraph_mapping(f, g, h):
            for v in g:
                if not h.has_node(f(v)):
                    raise RuntimeError(f"node {v} mapped to {f(v)} is not in {h.graph['name']} ({h.graph['labels']})")
            for u, v in g.edges:
                if not h.has_edge(f(u), f(v)):
                    raise RuntimeError(f"edge {(u, v)} mapped to {(f(u), f(v))} not present in {h.graph['name']} ({h.graph['labels']})")

        coords5 = dnx.pegasus_coordinates(5)
        coords3 = dnx.pegasus_coordinates(3)

        p3l = dnx.pegasus_graph(3)
        p3c = dnx.pegasus_graph(3, coordinates=True)
        p3n = coords3.graph_to_nice(p3c)

        p5l = dnx.pegasus_graph(5)
        p5c = dnx.pegasus_graph(5, coordinates=True)
        p5n = coords5.graph_to_nice(p5c)

        for target in p5l, p5c, p5n:
            for source in p3l, p3c, p3n:
                covered = set()
                for f in dnx.pegasus_sublattice_mappings(source, target):
                    check_subgraph_mapping(f, source, target)
                    covered.update(map(f, source))
                self.assertEqual(covered, set(target))

        c2l = dnx.chimera_graph(2)
        c2c = dnx.chimera_graph(2, coordinates=True)
        c23l = dnx.chimera_graph(2, 3)
        c32c = dnx.chimera_graph(3, 2, coordinates=True)

        p5n = dnx.pegasus_graph(5, nice_coordinates=True)
        p5l = coords5.graph_to_linear(p5n)
        p5c = coords5.graph_to_pegasus(p5n)

        for target in p5l, p5c, p5n:
            for source in c2l, c2c, c23l, c32c, target:
                covered = set()
                for f in dnx.pegasus_sublattice_mappings(source, target):
                    check_subgraph_mapping(f, source, target)
                    covered.update(map(f, source))
                self.assertEqual(covered, set(target))


class TestTupleFragmentation(unittest.TestCase):

    def test_empty_list(self):
        # Set up fragmentation function
        pg = dnx.pegasus_graph(3)
        fragment_tuple = get_tuple_fragmentation_fn(pg)

        # Fragment pegasus coordinates
        fragments = fragment_tuple([])
        self.assertEqual([], fragments)

    def test_single_horizontal_coordinate(self):
        # Set up fragmentation function
        pg = dnx.pegasus_graph(2)
        fragment_tuple = get_tuple_fragmentation_fn(pg)

        # Fragment pegasus coordinates
        pegasus_coord = (1, 0, 0, 0)
        fragments = fragment_tuple([pegasus_coord])

        expected_fragments = {(0, 3, 1, 0),
                              (0, 4, 1, 0),
                              (0, 5, 1, 0),
                              (0, 6, 1, 0),
                              (0, 7, 1, 0),
                              (0, 8, 1, 0)}

        self.assertEqual(expected_fragments, set(fragments))

    def test_single_vertical_coordinate(self):
        # Set up fragmentation function
        pg = dnx.pegasus_graph(6)
        fragment_tuple = get_tuple_fragmentation_fn(pg)

        pegasus_coord = (0, 1, 3, 1)
        fragments = fragment_tuple([pegasus_coord])

        expected_fragments = {(7, 7, 0, 1),
                              (8, 7, 0, 1),
                              (9, 7, 0, 1),
                              (10, 7, 0, 1),
                              (11, 7, 0, 1),
                              (12, 7, 0, 1)}

        self.assertEqual(expected_fragments, set(fragments))

    def test_list_of_coordinates(self):
        # Set up fragmentation function
        pg = dnx.pegasus_graph(6)
        fragment_tuple = get_tuple_fragmentation_fn(pg)

        # Fragment pegasus coordinates
        pegasus_coords = [(1, 5, 11, 4), (0, 2, 2, 3)]
        fragments = fragment_tuple(pegasus_coords)

        expected_fragments = {(35, 29, 1, 1),
                              (35, 30, 1, 1),
                              (35, 31, 1, 1),
                              (35, 32, 1, 1),
                              (35, 33, 1, 1),
                              (35, 34, 1, 1),
                              (19, 13, 0, 0),
                              (20, 13, 0, 0),
                              (21, 13, 0, 0),
                              (22, 13, 0, 0),
                              (23, 13, 0, 0),
                              (24, 13, 0, 0)}

        self.assertEqual(expected_fragments, set(fragments))


class TestTupleDefragmentation(unittest.TestCase):
    def test_empty_list(self):
        # Set up defragmentation function
        pg = dnx.pegasus_graph(2)
        defragment_tuple = get_tuple_defragmentation_fn(pg)

        # De-fragment chimera coordinates
        chimera_coords = []
        pegasus_coords = defragment_tuple(chimera_coords)

        self.assertEqual([], pegasus_coords)

    def test_single_fragment(self):
        # Set up defragmentation function
        pg = dnx.pegasus_graph(4)
        defragment_tuple = get_tuple_defragmentation_fn(pg)

        # De-fragment chimera coordinates
        chimera_coords = [(3, 7, 0, 0)]
        pegasus_coords = defragment_tuple(chimera_coords)

        expected_pegasus_coords = [(0, 1, 2, 0)]
        self.assertEqual(expected_pegasus_coords, pegasus_coords)

    def test_multiple_fragments_from_same_qubit(self):
        # Set up defragmentation function
        pg = dnx.pegasus_graph(3)
        defragment_tuple = get_tuple_defragmentation_fn(pg)

        # De-fragment chimera coordinates
        chimera_coords = [(9, 8, 1, 1), (9, 11, 1, 1)]
        pegasus_coords = defragment_tuple(chimera_coords)

        expected_pegasus_coords = [(1, 1, 7, 1)]
        self.assertEqual(expected_pegasus_coords, pegasus_coords)

    def test_mixed_fragments(self):
        # Set up defragmenation function
        pg = dnx.pegasus_graph(8)
        defragment_tuple = get_tuple_defragmentation_fn(pg)

        # De-fragment chimera coordinates
        chimera_coords = [(17, 14, 0, 0), (22, 14, 0, 0), (24, 32, 1, 0), (1, 31, 0, 0)]
        pegasus_coords = defragment_tuple(chimera_coords)

        expected_pegasus_coords = {(0, 2, 4, 2), (1, 4, 0, 4), (0, 5, 2, 0)}
        self.assertEqual(expected_pegasus_coords, set(pegasus_coords))

class TestFragmentedEdges(unittest.TestCase):
    def test_linear_indices(self):
        p = dnx.pegasus_graph(3, coordinates=False)
        c = dnx.chimera_graph(24, coordinates=True)
        num_edges = 0
        for u, v in fragmented_edges(p):
            self.assertTrue(c.has_edge(u, v))
            num_edges += 1
        #This is a weird edgecount: each node produces 5 extra edges for the internal connections
        #between fragments corresponding to a pegasus qubit.  But then we need to delete the odd
        #couplers, which aren't included in the chimera graph -- odd couplers make a perfect
        #matching, so thats 1/2 an edge per node.
        self.assertEqual(p.number_of_edges() + 9 * p.number_of_nodes()//2, num_edges)

    def test_coordinates(self):
        p = dnx.pegasus_graph(3, coordinates=True)
        c = dnx.chimera_graph(24, coordinates=True)
        num_edges = 0
        for u, v in fragmented_edges(p):
            self.assertTrue(c.has_edge(u, v))
            num_edges += 1

        #This is a weird edgecount: each node produces 5 extra edges for the internal connections
        #between fragments corresponding to a pegasus qubit.  But then we need to delete the odd
        #couplers, which aren't included in the chimera graph -- odd couplers make a perfect
        #matching, so thats 1/2 an edge per node.
        self.assertEqual(p.number_of_edges() + 9 * p.number_of_nodes()//2, num_edges)

    def test_nice_coordinates(self):
        p = dnx.pegasus_graph(3, nice_coordinates=True)
        c = dnx.chimera_graph(24, coordinates=True)
        num_edges = 0
        for u, v in fragmented_edges(p):
            self.assertTrue(c.has_edge(u, v))
            num_edges += 1
        #This is a weird edgecount: each node produces 5 extra edges for the internal connections
        #between fragments corresponding to a pegasus qubit.  But then we need to delete the odd
        #couplers, which aren't included in the chimera graph -- odd couplers make a perfect
        #matching, so thats 1/2 an edge per node.
        self.assertEqual(p.number_of_edges() + 9 * p.number_of_nodes()//2, num_edges)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import unittest

import networkx as nx

import dimod

import dwave_networkx as dnx


# adapted from future networkx version (nx.path_weight(...))
def path_weight(G, path):
    '''Return the total cost of a cycle in G specified by path.'''
    cost = 0
    for node, nbr in nx.utils.pairwise(path):
        cost += G[node][nbr]['weight']
    # add back to the starting point
    cost += G[path[-1]][path[0]]['weight']
    return cost


class TestIsHamiltonCycle(unittest.TestCase):
    def test_empty(self):
        G = nx.Graph()

        self.assertTrue(dnx.is_hamiltonian_path(G, []))

    def test_K1(self):
        G = nx.complete_graph(1)

        self.assertTrue(dnx.is_hamiltonian_path(G, [0]))
        self.assertFalse(dnx.is_hamiltonian_path(G, []))

    def test_K2(self):
        G = nx.complete_graph(2)

        self.assertTrue(dnx.is_hamiltonian_path(G, [0, 1]))
        self.assertTrue(dnx.is_hamiltonian_path(G, [1, 0]))
        self.assertFalse(dnx.is_hamiltonian_path(G, [0]))
        self.assertFalse(dnx.is_hamiltonian_path(G, [1]))
        self.assertFalse(dnx.is_hamiltonian_path(G, []))

    def test_K3(self):
        G = nx.complete_graph(3)

        self.assertTrue(dnx.is_hamiltonian_path(G, [0, 1, 2]))
        self.assertTrue(dnx.is_hamiltonian_path(G, [1, 0, 2]))
        self.assertFalse(dnx.is_hamiltonian_path(G, [0, 1]))
        self.assertFalse(dnx.is_hamiltonian_path(G, [0]))
        self.assertFalse(dnx.is_hamiltonian_path(G, [1]))
        self.assertFalse(dnx.is_hamiltonian_path(G, []))


class TestTSP(unittest.TestCase):

    def test_TSP_basic(self):
        """Runs the function on some small and simple graphs, just to make
        sure it works in basic functionality.
        """
        G = nx.complete_graph(4)
        for u, v in G.edges():
            G[u][v]['weight'] = 1
        route = dnx.traveling_salesperson(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_hamiltonian_path(G, route))

        G = nx.complete_graph(4)
        for u, v in G.edges():
            G[u][v]['weight'] = u+v
        route = dnx.traveling_salesperson(G, dimod.ExactSolver(), lagrange=10.0)
        self.assertTrue(dnx.is_hamiltonian_path(G, route))

    def test_dimod_vs_list(self):
        G = nx.complete_graph(4)
        for u, v in G.edges():
            G[u][v]['weight'] = 1

        route = dnx.traveling_salesperson(G, dimod.ExactSolver())
        route = dnx.traveling_salesperson(G, dimod.SimulatedAnnealingSampler())

    def test_weighted_complete_graph(self):
        G = nx.Graph()
        G.add_weighted_edges_from({(0, 1, 1), (0, 2, 2), (0, 3, 3), (1, 2, 3),
                                   (1, 3, 4), (2, 3, 5)})
        route = dnx.traveling_salesperson(G, dimod.ExactSolver(), lagrange=10)

        self.assertEqual(len(route), len(G))

    def test_start(self):
        G = nx.Graph()
        G.add_weighted_edges_from((u, v, .5)
                                  for u, v in itertools.combinations(range(3), 2))

        route = dnx.traveling_salesperson(G, dimod.ExactSolver(), start=1)

        self.assertEqual(route[0], 1)

    def test_weighted_complete_digraph(self):
        G = nx.DiGraph()
        G.add_weighted_edges_from([
            (0, 1, 2),
            (1, 0, 1),
            (0, 2, 2),
            (2, 0, 2),
            (0, 3, 1),
            (3, 0, 2),
            (1, 2, 2),
            (2, 1, 1),
            (1, 3, 2),
            (3, 1, 2),
            (2, 3, 2),
            (3, 2, 1),
        ])

        route = dnx.traveling_salesperson(G, dimod.ExactSolver(), start=1)

        self.assertEqual(len(route), len(G))
        self.assertListEqual(route, [1, 0, 3, 2])

        cost = path_weight(G, route)

        self.assertEqual(cost, 4)


class TestTSPQUBO(unittest.TestCase):
    def test_empty(self):
        Q = dnx.traveling_salesperson_qubo(nx.Graph())
        self.assertEqual(Q, {})

    def test_k3(self):
        # 3cycle so all paths are equally good
        G = nx.Graph()
        G.add_weighted_edges_from([('a', 'b', 0.5),
                                   ('b', 'c', 1.0),
                                   ('a', 'c', 2.0)])

        Q = dnx.traveling_salesperson_qubo(G, lagrange=10)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        # all routes are min weight
        min_routes = list(itertools.permutations(G.nodes))

        # get the min energy of the qubo
        sampleset = dimod.ExactSolver().sample(bqm)
        ground_energy = sampleset.first.energy

        # all possible routes are equally good
        for route in min_routes:
            sample = {v: 0 for v in bqm.variables}
            for idx, city in enumerate(route):
                sample[(city, idx)] = 1
            self.assertAlmostEqual(bqm.energy(sample), ground_energy)

        # all min-energy solutions are valid routes
        ground_count = 0
        for sample, energy in sampleset.data(['sample', 'energy']):
            if abs(energy - ground_energy) > .001:
                break
            ground_count += 1

        self.assertEqual(ground_count, len(min_routes))

    def test_k3_bidirectional(self):
        G = nx.DiGraph()
        G.add_weighted_edges_from([('a', 'b', 0.5),
                                   ('b', 'a', 0.5),
                                   ('b', 'c', 1.0),
                                   ('c', 'b', 1.0),
                                   ('a', 'c', 2.0),
                                   ('c', 'a', 2.0)])

        Q = dnx.traveling_salesperson_qubo(G, lagrange=10)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        # all routes are min weight
        min_routes = list(itertools.permutations(G.nodes))

        # get the min energy of the qubo
        sampleset = dimod.ExactSolver().sample(bqm)
        ground_energy = sampleset.first.energy

        # all possible routes are equally good
        for route in min_routes:
            sample = {v: 0 for v in bqm.variables}
            for idx, city in enumerate(route):
                sample[(city, idx)] = 1
            self.assertAlmostEqual(bqm.energy(sample), ground_energy)

        # all min-energy solutions are valid routes
        ground_count = 0
        for sample, energy in sampleset.data(['sample', 'energy']):
            if abs(energy - ground_energy) > .001:
                break
            ground_count += 1

        self.assertEqual(ground_count, len(min_routes))

    def test_graph_missing_edges(self):
        G1 = nx.Graph()
        G1.add_weighted_edges_from([
            ('a', 'b', 0.5),
            ('b', 'c', 1.0),
            ('a', 'c', 2.0),
        ])
        Q1 = dnx.traveling_salesperson_qubo(G1, lagrange=10)

        G2 = nx.Graph()
        G2.add_weighted_edges_from([
            ('a', 'b', 0.5),
            ('a', 'c', 2.0),
        ])
        # make sure that missing_edge_weight gets applied correctly
        Q2 = dnx.traveling_salesperson_qubo(G2, lagrange=10, missing_edge_weight=1.0)

        self.assertDictEqual(Q1, Q2)

    def test_digraph_missing_edges(self):
        G1 = nx.DiGraph()
        G1.add_weighted_edges_from([
            ('a', 'b', 0.5),
            ('b', 'a', 0.8),
            ('b', 'c', 1.0),
            ('c', 'b', 0.7),
            ('a', 'c', 2.0),
            ('c', 'a', 2.0),
        ])
        Q1 = dnx.traveling_salesperson_qubo(G1, lagrange=10)

        G2 = nx.DiGraph()
        G2.add_weighted_edges_from([
            ('a', 'b', 0.5),
            ('b', 'a', 0.8),
            ('c', 'b', 0.7),
            ('a', 'c', 2.0),
            ('c', 'a', 2.0),
        ])

        # make sure that missing_edge_weight gets applied correctly
        Q2 = dnx.traveling_salesperson_qubo(G2, lagrange=10, missing_edge_weight=1.0)

        self.assertDictEqual(Q1, Q2)

    def test_k4_equal_weights(self):
        # k5 with all equal weights so all paths are equally good
        G = nx.Graph()
        G.add_weighted_edges_from((u, v, .5)
                                  for u, v in itertools.combinations(range(4), 2))

        Q = dnx.traveling_salesperson_qubo(G, lagrange=10)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        # all routes are min weight
        min_routes = list(itertools.permutations(G.nodes))

        # get the min energy of the qubo
        sampleset = dimod.ExactSolver().sample(bqm)
        ground_energy = sampleset.first.energy

        # all possible routes are equally good
        for route in min_routes:
            sample = {v: 0 for v in bqm.variables}
            for idx, city in enumerate(route):
                sample[(city, idx)] = 1
            self.assertAlmostEqual(bqm.energy(sample), ground_energy)

        # all min-energy solutions are valid routes
        ground_count = 0
        for sample, energy in sampleset.data(['sample', 'energy']):
            if abs(energy - ground_energy) > .001:
                break
            ground_count += 1

        self.assertEqual(ground_count, len(min_routes))

    def test_k4(self):
        # good routes are 0,1,2,3 or 3,2,1,0 (and their rotations)
        G = nx.Graph()
        G.add_weighted_edges_from([(0, 1, 1),
                                   (1, 2, 1),
                                   (2, 3, 1),
                                   (3, 0, 1),
                                   (0, 2, 2),
                                   (1, 3, 2)])

        Q = dnx.traveling_salesperson_qubo(G, lagrange=10)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        # good routes won't have 0<->2 or 1<->3
        min_routes = [(0, 1, 2, 3),
                      (1, 2, 3, 0),
                      (2, 3, 0, 1),
                      (1, 2, 3, 0),
                      (3, 2, 1, 0),
                      (2, 1, 0, 3),
                      (1, 0, 3, 2),
                      (0, 3, 2, 1)]

        # get the min energy of the qubo
        sampleset = dimod.ExactSolver().sample(bqm)
        ground_energy = sampleset.first.energy

        # all possible routes are equally good
        for route in min_routes:
            sample = {v: 0 for v in bqm.variables}
            for idx, city in enumerate(route):
                sample[(city, idx)] = 1
            self.assertAlmostEqual(bqm.energy(sample), ground_energy)

        # all min-energy solutions are valid routes
        ground_count = 0
        for sample, energy in sampleset.data(['sample', 'energy']):
            if abs(energy - ground_energy) > .001:
                break
            ground_count += 1

        self.assertEqual(ground_count, len(min_routes))

    def test_weighted_complete_graph(self):
        G = nx.Graph()
        G.add_weighted_edges_from({(0, 1, 1), (0, 2, 100),
                                   (0, 3, 1), (1, 2, 1),
                                   (1, 3, 100), (2, 3, 1)})

        lagrange = 5.0

        Q = dnx.traveling_salesperson_qubo(G, lagrange, 'weight')

        N = G.number_of_nodes()
        correct_sum = G.size('weight')*2*N-2*N*N*lagrange+2*N*N*(N-1)*lagrange

        actual_sum = sum(Q.values())

        self.assertEqual(correct_sum, actual_sum)

    def test_exceptions(self):
        G = nx.Graph([(0, 1)])
        with self.assertRaises(ValueError):
            dnx.traveling_salesperson_qubo(G)

    def test_docstring_size(self):
        # in the docstring we state the size of the resulting BQM, this checks
        # that
        for n in range(3, 20):
            G = nx.Graph()
            G.add_weighted_edges_from((u, v, .5)
                                      for u, v
                                      in itertools.combinations(range(n), 2))
            Q = dnx.traveling_salesperson_qubo(G)
            bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

            self.assertEqual(len(bqm), n**2)
            self.assertEqual(len(bqm.quadratic), 2*n*n*(n - 1))



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import itertools

import networkx as nx
import dwave_networkx as dnx
from dimod import ExactSolver, SimulatedAnnealingSampler


class TestSocial(unittest.TestCase):
    def check_bicolor(self, colors):
        # colors should be ints and either 0 or 1
        for c in colors.values():
            self.assertTrue(c in (0, 1))

    def test_structural_imbalance_basic(self):
        sampler = ExactSolver()

        blueteam = ['Alice', 'Bob', 'Carol']
        redteam0 = ['Eve']
        redteam1 = ['Mallory', 'Trudy']

        S = nx.Graph()
        for p0, p1 in itertools.combinations(blueteam, 2):
            S.add_edge(p0, p1, sign=1)
        S.add_edge(*redteam1, sign=1)
        for p0 in blueteam:
            for p1 in redteam0:
                S.add_edge(p0, p1, sign=-1)
            for p1 in redteam1:
                S.add_edge(p0, p1, sign=-1)

        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
        self.check_bicolor(colors)

        greenteam = ['Ted']
        for p0 in set(S.nodes):
            for p1 in greenteam:
                S.add_edge(p0, p1, sign=1)

        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
        self.check_bicolor(colors)

    def test_structural_imbalance_docstring_example(self):
        sampler = ExactSolver()

        S = nx.Graph()
        S.add_edge('Alice', 'Bob', sign=1)  # Alice and Bob are friendly
        S.add_edge('Alice', 'Eve', sign=-1)  # Alice and Eve are hostile
        S.add_edge('Bob', 'Eve', sign=-1)  # Bob and Eve are hostile
        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
        self.check_bicolor(colors)
        self.assertEqual(frustrated_edges, {})
        S.add_edge('Ted', 'Bob', sign=1)  # Ted is friendly with all
        S.add_edge('Ted', 'Alice', sign=1)
        S.add_edge('Ted', 'Eve', sign=1)
        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
        self.check_bicolor(colors)
        self.assertTrue(frustrated_edges == {('Ted', 'Eve'): {'sign': 1}} or
                        frustrated_edges == {('Eve', 'Ted'): {'sign': 1}})

    def test_default_sampler(self):
        S = nx.Graph()
        S.add_edge('Alice', 'Bob', sign=1)  # Alice and Bob are friendly
        S.add_edge('Alice', 'Eve', sign=-1)  # Alice and Eve are hostile
        S.add_edge('Bob', 'Eve', sign=-1)  # Bob and Eve are hostile

        dnx.set_default_sampler(ExactSolver())
        self.assertIsNot(dnx.get_default_sampler(), None)
        frustrated_edges, colors = dnx.structural_imbalance(S)
        dnx.unset_default_sampler()
        self.assertEqual(dnx.get_default_sampler(), None, "sampler did not unset correctly")

    def test_invalid_graph(self):
        """should throw an error with a graph without sign attribute"""
        S = nx.Graph()
        S.add_edge('Alice', 'Bob', sign=1)
        S.add_edge('Alice', 'Eve', sign=-1)
        S.add_edge('Bob', 'Eve')  # invalid edge

        with self.assertRaises(ValueError):
            frustrated_edges, colors = dnx.structural_imbalance(S, ExactSolver())

    def test_sign_zero(self):
        """though not documentented, agents with no relation can have sign 0.
        This is less performant than just not having an edge in most cases."""
        sampler = ExactSolver()

        S = nx.Graph()
        S.add_edge('Alice', 'Bob', sign=1)  # Alice and Bob are friendly
        S.add_edge('Alice', 'Eve', sign=-1)  # Alice and Eve are hostile
        S.add_edge('Bob', 'Eve', sign=0)  # Bob and Eve have no relation

        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)

        self.assertEqual(frustrated_edges, {})  # should be no frustration

    def test_frustrated_hostile_edge(self):
        """Set up a graph where the frustrated edge should be hostile"""
        sampler = ExactSolver()

        S = nx.florentine_families_graph()

        # set all hostile
        nx.set_edge_attributes(S, -1, 'sign')

        # smoke test
        frustrated_edges, colors = dnx.structural_imbalance(S, sampler)
        self.check_bicolor(colors)



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from dwave_networkx.generators.chimera import chimera_coordinates

__all__ = ['canonical_chimera_labeling']


def canonical_chimera_labeling(G, t=None):
    """Returns a mapping from the labels of G to chimera-indexed labeling.

    Parameters
    ----------
    G : NetworkX graph
        A Chimera-structured graph.
    t : int (optional, default 4)
        Size of the shore within each Chimera tile.

    Returns
    -------
    chimera_indices: dict
        A mapping from the current labels to a 4-tuple of Chimera indices.

    """
    adj = G.adj

    if t is None:
        if hasattr(G, 'edges'):
            num_edges = len(G.edges)
        else:
            num_edges = len(G.quadratic)
        t = _chimera_shore_size(adj, num_edges)

    chimera_indices = {}

    row = col = 0

    # need to find a node in a corner
    root_edge = min(((u, v) for u in adj for v in adj[u]),
                    key=lambda edge: len(adj[edge[0]]) + len(adj[edge[1]]))
    root, _ = root_edge

    horiz, verti = rooted_tile(adj, root, t)
    while len(chimera_indices) < len(adj):

        new_indices = {}

        if row == 0:
            # if we're in the 0th row, we can assign the horizontal randomly
            for si, v in enumerate(horiz):
                new_indices[v] = (row, col, 0, si)
        else:
            # we need to match the row above
            for v in horiz:
                north = [u for u in adj[v] if u in chimera_indices]
                assert len(north) == 1
                i, j, u, si = chimera_indices[north[0]]
                assert i == row - 1 and j == col and u == 0
                new_indices[v] = (row, col, 0, si)

        if col == 0:
            # if we're in the 0th col, we can assign the vertical randomly
            for si, v in enumerate(verti):
                new_indices[v] = (row, col, 1, si)
        else:
            # we need to match the column to the east
            for v in verti:
                east = [u for u in adj[v] if u in chimera_indices]
                assert len(east) == 1
                i, j, u, si = chimera_indices[east[0]]
                assert i == row and j == col - 1 and u == 1
                new_indices[v] = (row, col, 1, si)

        chimera_indices.update(new_indices)

        # get the next root
        root_neighbours = [v for v in adj[root] if v not in chimera_indices]
        if len(root_neighbours) == 1:
            # we can increment the row
            root = root_neighbours[0]
            horiz, verti = rooted_tile(adj, root, t)

            row += 1
        else:
            # need to go back to row 0, and increment the column
            assert not root_neighbours  # should be empty

            # we want (0, col, 1, 0), we could cache this, but for now let's just go look for it
            # the slow way
            vert_root = [v for v in chimera_indices if chimera_indices[v] == (0, col, 1, 0)][0]

            vert_root_neighbours = [v for v in adj[vert_root] if v not in chimera_indices]

            if vert_root_neighbours:

                verti, horiz = rooted_tile(adj, vert_root_neighbours[0], t)
                root = next(iter(horiz))

                row = 0
                col += 1

    return chimera_indices


def rooted_tile(adj, n, t):
    horiz = {n}
    vert = set()

    # get all of the nodes that are two steps away from n
    two_steps = {v for u in adj[n] for v in adj[u] if v != n}

    # find the subset of two_steps that share exactly t neighbours
    for v in two_steps:
        shared = set(adj[n]).intersection(adj[v])

        if len(shared) == t:
            assert v not in horiz
            horiz.add(v)
            vert |= shared

    assert len(vert) == t
    return horiz, vert


def _chimera_shore_size(adj, num_edges):
    # we know |E| = m*n*t*t + (2*m*n-m-n)*t

    num_nodes = len(adj)

    max_degree = max(len(adj[v]) for v in adj)

    if num_nodes == 2 * max_degree:
        return max_degree

    def a(t):
        return -2*t

    def b(t):
        return (t + 2) * num_nodes - 2 * num_edges

    def c(t):
        return -num_nodes

    t = max_degree - 1
    m = (-b(t) + math.sqrt(b(t)**2 - 4*a(t)*c(t))) / (2 * a(t))

    if m.is_integer():
        return t

    return max_degree - 2



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import dwave_networkx as dnx
import networkx as nx
import dimod

# Use basic simulated annealer
sampler = dimod.SimulatedAnnealingSampler()

# Set up a Networkx Graph
G = nx.Graph()
G.add_edges_from([(0, 1), (0, 2), (1, 2), (1, 3), (2, 3), (1, 4), (2, 4), (3, 4), (3, 5), (4, 5), (5, 2)])

# Get the minimum vertex coloring, which is known in this case to be of
# length 6
candidate = dnx.min_vertex_coloring(G, sampler)
if dnx.is_vertex_coloring(G, candidate) and len(candidate) == 6:
    print(candidate, " is a minimum vertex coloring")
else:
    print(candidate, " is not a minimum vertex coloring")



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import collections

import dimod
import dwave_networkx as dnx
import networkx as nx


class TestRootedTile(unittest.TestCase):
    def test_C33_tiles(self):
        # not a public function
        rooted_tile = dnx.algorithms.canonicalization.rooted_tile

        C33 = dnx.chimera_graph(3, 3, 4)

        for root in range(0, len(C33), 8):

            horiz, vert = rooted_tile(C33.adj, root, 4)

            self.assertEqual(horiz, set(range(root, root+4)))
            self.assertEqual(vert, set(range(root+4, root+8)))


class TestCanonicalChimeraLabeling(unittest.TestCase):
    def test_tile_identity(self):
        C1 = dnx.chimera_graph(1)
        coord = dnx.chimera_coordinates(1, 1, 4)

        labels = dnx.canonical_chimera_labeling(C1)
        labels = {v: coord.chimera_to_linear(labels[v]) for v in labels}

        G = nx.relabel_nodes(C1, labels, copy=True)

        self.assertTrue(nx.is_isomorphic(G, C1))
        self.assertEqual(set(G), set(C1))

    def test_bqm_tile_identity(self):
        J = {e: -1 for e in dnx.chimera_graph(1).edges}
        C1bqm = dimod.BinaryQuadraticModel.from_ising({}, J)
        coord = dnx.chimera_coordinates(1, 1, 4)

        labels = dnx.canonical_chimera_labeling(C1bqm)
        labels = {v: coord.chimera_to_linear(labels[v]) for v in labels}

        bqm = C1bqm.relabel_variables(labels, inplace=False)

        self.assertEqual(bqm, C1bqm)

    def test_row_identity(self):
        C41 = dnx.chimera_graph(4, 1)
        coord = dnx.chimera_coordinates(4, 1, 4)

        labels = dnx.canonical_chimera_labeling(C41)
        labels = {v: coord.chimera_to_linear(labels[v]) for v in labels}

        G = nx.relabel_nodes(C41, labels, copy=True)

        self.assertTrue(nx.is_isomorphic(G, C41))

    def test_3x3_identity(self):
        C33 = dnx.chimera_graph(3, 3)
        coord = dnx.chimera_coordinates(3, 3, 4)

        labels = dnx.canonical_chimera_labeling(C33)
        labels = {v: coord.chimera_to_linear(labels[v]) for v in labels}

        G = nx.relabel_nodes(C33, labels, copy=True)

        self.assertTrue(nx.is_isomorphic(G, C33))

    def test_construction_string_labels(self):
        C22 = dnx.chimera_graph(2, 2, 3)
        coord = dnx.chimera_coordinates(2, 2, 3)

        alpha = 'abcdefghijklmnopqrstuvwxyz'

        bqm = dimod.BinaryQuadraticModel.empty(dimod.BINARY)

        for u, v in reversed(list(C22.edges)):
            bqm.add_interaction(alpha[u], alpha[v], 1)

        assert len(bqm.quadratic) == len(C22.edges)
        assert len(bqm) == len(C22)

        labels = dnx.canonical_chimera_labeling(bqm)
        labels = {v: alpha[coord.chimera_to_linear(labels[v])] for v in labels}

        bqm2 = bqm.relabel_variables(labels, inplace=False)

        self.assertEqual(bqm, bqm2)

    def test_reversed(self):
        C33 = nx.OrderedGraph()
        C33.add_nodes_from(reversed(range(3*3*4)))
        C33.add_edges_from(dnx.chimera_graph(3, 3, 4).edges)
        coord = dnx.chimera_coordinates(3, 3, 4)

        labels = dnx.canonical_chimera_labeling(C33)
        labels = {v: coord.chimera_to_linear(labels[v]) for v in labels}

        G = nx.relabel_nodes(C33, labels, copy=True)

        self.assertTrue(nx.is_isomorphic(G, C33))

    def test__shore_size_tiles(self):
        shore_size = dnx.algorithms.canonicalization._chimera_shore_size

        for t in range(1, 8):
            G = dnx.chimera_graph(1, 1, t)
            self.assertEqual(shore_size(G.adj, len(G.edges)), t)

    def test__shore_size_columns(self):
        shore_size = dnx.algorithms.canonicalization._chimera_shore_size

        # 2, 1, 1 is the same as 1, 1, 2
        for m in range(2, 11):
            for t in range(9, 1, -1):
                G = dnx.chimera_graph(m, 1, t)
                self.assertEqual(shore_size(G.adj, len(G.edges)), t)

    def test__shore_size_rectangles(self):
        shore_size = dnx.algorithms.canonicalization._chimera_shore_size

        # 2, 1, 1 is the same as 1, 1, 2
        for m in range(2, 7):
            for n in range(2, 7):
                for t in range(1, 6):
                    G = dnx.chimera_graph(m, n, t)
                    self.assertEqual(shore_size(G.adj, len(G.edges)), t)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dwave_networkx as dnx

class TestZephyrGraph(unittest.TestCase):
    def test_single_tile(self):

        # fully specified
        G = dnx.zephyr_graph(1, 4)

        # should have 8 nodes
        self.assertEqual(len(G), 48)

        # nodes 0,...,7 should be in the tile
        for n in range(48):
            self.assertIn(n, G)


    def test_not_full_yield(self):
        edges =  [(2, 30), (7, 44), (10, 37), (12, 29), (15, 37), (19, 41)]
        G = dnx.zephyr_graph(1, 4, edge_list=edges)
        for e in edges:
            self.assertIn(e, G.edges())
        for (u, v) in G.edges:
            self.assertTrue((u, v) in edges or (v, u) in edges)

        nodes = [0, 1, 2]
        G = dnx.zephyr_graph(1, 2, node_list=nodes)
        self.assertEqual(len(G), 3)
        self.assertEqual(len(G.edges()), 1)

        edges = [(0, 1), (2, 3)]
        nodes = [0, 1, 2, 3]
        G = dnx.zephyr_graph(1, 2, node_list=nodes, edge_list=edges)
        self.assertEqual(len(G), 4)
        self.assertEqual(len(G.edges()), 2)

    def test_float_robustness(self):
        G = dnx.zephyr_graph(8 / 2)

        self.assertEqual(set(G.nodes), set(dnx.zephyr_graph(4).nodes))
        for u, v in dnx.zephyr_graph(4).edges:
            self.assertIn(u, G[v])

        G = dnx.zephyr_graph(4, 4.)

        self.assertEqual(set(G.nodes), set(dnx.zephyr_graph(4).nodes))
        for u, v in dnx.zephyr_graph(4).edges:
            self.assertIn(u, G[v])

    def test_coordinate_basics(self):
        from dwave_networkx.generators.zephyr import zephyr_coordinates
        G = dnx.zephyr_graph(4)
        H = dnx.zephyr_graph(4, coordinates=True)
        coords = zephyr_coordinates(4)
        Gnodes = G.nodes
        Hnodes = H.nodes
        for v in Gnodes:
            q = Gnodes[v]['zephyr_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.zephyr_to_linear(q))
            self.assertEqual(q, coords.linear_to_zephyr(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['zephyr_index'], q)
            self.assertEqual(v, coords.zephyr_to_linear(q))
            self.assertEqual(q, coords.linear_to_zephyr(v))

    def test_coordinate_subgraphs(self):
        from dwave_networkx.generators.zephyr import zephyr_coordinates
        from random import sample
        G = dnx.zephyr_graph(4)
        H = dnx.zephyr_graph(4, coordinates=True)
        coords = zephyr_coordinates(4)

        lmask = sample(list(G.nodes()), G.number_of_nodes()//2)
        cmask = list(coords.iter_linear_to_zephyr(lmask))

        self.assertEqual(lmask, list(coords.iter_zephyr_to_linear(cmask)))

        Gm = dnx.zephyr_graph(4, node_list=lmask)
        Hm = dnx.zephyr_graph(4, node_list=cmask, coordinates=True)

        Gs = G.subgraph(lmask)
        Hs = H.subgraph(cmask)

        EG = sorted(map(sorted, Gs.edges()))
        EH = sorted(map(sorted, Hs.edges()))

        self.assertEqual(EG, sorted(map(sorted, Gm.edges())))
        self.assertEqual(EH, sorted(map(sorted, Hm.edges())))

        Gn = dnx.zephyr_graph(4, edge_list=EG)
        Hn = dnx.zephyr_graph(4, edge_list=EH, coordinates=True)

        Gnodes = Gn.nodes
        Hnodes = Hn.nodes
        for v in Gnodes:
            q = Gnodes[v]['zephyr_index']
            self.assertIn(q, Hnodes)
            self.assertEqual(Hnodes[q]['linear_index'], v)
            self.assertEqual(v, coords.zephyr_to_linear(q))
            self.assertEqual(q, coords.linear_to_zephyr(v))
        for q in Hnodes:
            v = Hnodes[q]['linear_index']
            self.assertIn(v, Gnodes)
            self.assertEqual(Gnodes[v]['zephyr_index'], q)
            self.assertEqual(v, coords.zephyr_to_linear(q))
            self.assertEqual(q, coords.linear_to_zephyr(v))

        self.assertEqual(EG, sorted(map(sorted, coords.iter_zephyr_to_linear_pairs(Hn.edges()))))
        self.assertEqual(EH, sorted(map(sorted, coords.iter_linear_to_zephyr_pairs(Gn.edges()))))

    def test_graph_relabeling(self):
        def graph_equal(g, h):
            self.assertEqual(set(g), set(h))
            self.assertEqual(
                set(map(tuple, map(sorted, g.edges))),
                set(map(tuple, map(sorted, g.edges)))
            )
            for v, d in g.nodes(data=True):
                self.assertEqual(h.nodes[v], d)

        coords = dnx.zephyr_coordinates(3)
        for data in True, False:
            z3l = dnx.zephyr_graph(3, data=data)
            z3c = dnx.zephyr_graph(3, data=data, coordinates=True)

            graph_equal(z3l, coords.graph_to_linear(z3l))
            graph_equal(z3l, coords.graph_to_linear(z3c))
            
            graph_equal(z3c, coords.graph_to_zephyr(z3c))
            graph_equal(z3c, coords.graph_to_zephyr(z3l))

        h = dnx.zephyr_graph(2)
        del h.graph['labels']
        with self.assertRaises(ValueError):
            coords.graph_to_linear(h)
        with self.assertRaises(ValueError):
            coords.graph_to_zephyr(h)


    def test_sublattice_mappings(self):
        def check_subgraph_mapping(f, g, h):
            for v in g:
                if not h.has_node(f(v)):
                    raise RuntimeError(f"node {v} mapped to {f(v)} is not in {h.graph['name']} ({h.graph['labels']})")
            for u, v in g.edges:
                if not h.has_edge(f(u), f(v)):
                    raise RuntimeError(f"edge {(u, v)} mapped to {(f(u), f(v))} not present in {h.graph['name']} ({h.graph['labels']})")

        z2l = dnx.zephyr_graph(2)
        z2c = dnx.zephyr_graph(2, coordinates=True)
        c2l = dnx.chimera_graph(2)
        c2c = dnx.chimera_graph(2, coordinates=True)
        c23l = dnx.chimera_graph(2, 3)
        c32c = dnx.chimera_graph(3, 2, coordinates=True)
        c2l8 = dnx.chimera_graph(2, t=8)
        c2c8 = dnx.chimera_graph(2, t=8, coordinates=True)        
        c23l8 = dnx.chimera_graph(2, 3, t=8)
        c32c8 = dnx.chimera_graph(3, 2, t=8, coordinates=True)        

        z5l = dnx.zephyr_graph(5)
        z5c = dnx.zephyr_graph(5, coordinates=True)

        for target in z5l, z5c:
            for source in z2l, z2c, c2l, c2c, c2l8, c2c8, c23l, c32c, c23l8, c32c8, target:
                covered = set()
                for f in dnx.zephyr_sublattice_mappings(source, target):
                    check_subgraph_mapping(f, source, target)
                    covered.update(map(f, source))
                self.assertEqual(covered, set(target))




# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Generators for graphs derived from the D-Wave System.

"""
import warnings

import networkx as nx
from networkx.algorithms.bipartite import color
from networkx import diameter

from dwave_networkx.exceptions import DWaveNetworkXException

from itertools import product

__all__ = ['chimera_graph',
           'chimera_coordinates',
           'find_chimera_indices',
           'chimera_to_linear',
           'linear_to_chimera',
           'chimera_sublattice_mappings',
           ]


def chimera_graph(m, n=None, t=None, create_using=None, node_list=None, edge_list=None, data=True, coordinates=False):
    """Creates a Chimera lattice of size (m, n, t).

    Parameters
    ----------
    m : int
        Number of rows in the Chimera lattice.
    n : int (optional, default m)
        Number of columns in the Chimera lattice.
    t : int (optional, default 4)
        Size of the shore within each Chimera tile.
    create_using : Graph (optional, default None)
        If provided, this graph is cleared of nodes and edges and filled
        with the new graph. Usually used to set the type of the graph.
    node_list : iterable (optional, default None)
        Iterable of nodes in the graph. If None, calculated
        from (m, n, t). Note that this list is used to remove nodes,
        so any nodes specified not in ``range(m * n * 2 * t)`` are not added.
    edge_list : iterable (optional, default None)
        Iterable of edges in the graph. If None, edges are
        generated as described below. The nodes in each edge must be
        integer-labeled in ``range(m * n * t * 2)``.
    data : bool (optional, default True)
        If True, each node has a
        `chimera_index attribute`. The attribute is a 4-tuple Chimera index
        as defined below.
    coordinates : bool (optional, default False)
        If True, node labels are 4-tuples, equivalent to the chimera_index
        attribute as below.  In this case, the `data` parameter controls the
        existence of a `linear_index attribute`, which is an int.

    Returns
    -------
    G : NetworkX Graph
        An (m, n, t) Chimera lattice. Nodes are labeled by integers.


    A Chimera lattice is an m-by-n grid of Chimera tiles. Each Chimera
    tile is itself a bipartite graph with shores of size t. The
    connection in a Chimera lattice can be expressed using a node-indexing
    notation (i,j,u,k) for each node.

    * (i,j) indexes the (row, column) of the Chimera tile. i must be
      between 0 and m-1, inclusive, and j must be between 0 and
      n-1, inclusive.
    * u=0 indicates the left-hand nodes in the tile, and u=1 indicates
      the right-hand nodes.
    * k=0,1,...,t-1 indexes nodes within either the left- or
      right-hand shores of a tile.

    In this notation, two nodes (i, j, u, k) and (i', j', u', k') are
    neighbors if and only if:

        (i = i' AND j = j' AND u != u') OR
        (i = i' +/- 1 AND j = j' AND u = 0 AND u' = 0 AND k = k') OR
        (i = i' AND j = j' +/- 1 AND u = 1 AND u' = 1 AND k = k')

    The first of the three terms of the disjunction gives the
    bipartite connections  within the tile. The second and third terms
    give the vertical and horizontal connections between blocks
    respectively.

    Node (i, j, u, k) is labeled by:

        label = i * n * 2 * t + j * 2 * t + u * t + k

    Examples
    ========
    >>> G = dnx.chimera_graph(1, 1, 2)  # a single Chimera tile
    >>> len(G)
    4
    >>> list(G.nodes())  # doctest: +SKIP
    [0, 1, 2, 3]
    >>> list(G.nodes(data=True))  # doctest: +SKIP
    [(0, {'chimera_index': (0, 0, 0, 0)}),
     (1, {'chimera_index': (0, 0, 0, 1)}),
     (2, {'chimera_index': (0, 0, 1, 0)}),
     (3, {'chimera_index': (0, 0, 1, 1)})]
    >>> list(G.edges())  # doctest: +SKIP
    [(0, 2), (0, 3), (1, 2), (1, 3)]

    """
    m = int(m)
    if n is None:
        n = m
    else:
        n = int(n)
    if t is None:
        t = 4
    else:
        t = int(t)

    G = nx.empty_graph(0, create_using)

    G.name = "chimera_graph(%s, %s, %s)" % (m, n, t)

    construction = (("family", "chimera"), ("rows", m), ("columns", n),
                    ("tile", t), ("data", data),
                    ("labels", "coordinate" if coordinates else "int"))

    G.graph.update(construction)

    max_size = m * n * 2 * t  # max number of nodes G can have

    if edge_list is None:
        if coordinates:
            # tile edges
            G.add_edges_from(((i, j, 0, k0), (i, j, 1, k1))
                             for i in range(m)
                             for j in range(n)
                             for k0 in range(t)
                             for k1 in range(t))

            # horizontal edges
            G.add_edges_from(((i, j, 1, k), (i, j+1, 1, k))
                             for i in range(m)
                             for j in range(n-1)
                             for k in range(t))
            # vertical edges
            G.add_edges_from(((i, j, 0, k), (i+1, j, 0, k))
                             for i in range(m-1)
                             for j in range(n)
                             for k in range(t))
        else:
            hoff = 2 * t
            voff = n * hoff
            mi = m * voff
            ni = n * hoff

            # tile edges
            G.add_edges_from((k0, k1)
                             for i in range(0, ni, hoff)
                             for j in range(i, mi, voff)
                             for k0 in range(j, j + t)
                             for k1 in range(j + t, j + 2 * t))
            # horizontal edges
            G.add_edges_from((k, k + hoff)
                             for i in range(t, 2 * t)
                             for j in range(i, ni - hoff, hoff)
                             for k in range(j, mi, voff))
            # vertical edges
            G.add_edges_from((k, k + voff)
                             for i in range(t)
                             for j in range(i, ni, hoff)
                             for k in range(j, mi - voff, voff))
    else:
        G.add_edges_from(edge_list)

    if node_list is not None:
        nodes = set(node_list)
        G.remove_nodes_from(set(G) - nodes)
        G.add_nodes_from(nodes)  # for singleton nodes

    if data:
        if coordinates:
            def checkadd(v, q):
                if q in G:
                    G.nodes[q]['linear_index'] = v
        else:
            def checkadd(v, q):
                if v in G:
                    G.nodes[v]['chimera_index'] = q

        v = 0
        for i in range(m):
            for j in range(n):
                for u in range(2):
                    for k in range(t):
                        checkadd(v, (i, j, u, k))
                        v += 1

    return G


def find_chimera_indices(G):
    """Attempts to determine the Chimera indices of the nodes in graph G.

    See the :func:`~chimera_graph()` function for a definition of a Chimera graph and Chimera
    indices.

    Parameters
    ----------
    G : NetworkX graph
        Should be a single-tile Chimera graph.

    Returns
    -------
    chimera_indices : dict
        A dict of the form {node: (i, j, u, k), ...} where (i, j, u, k)
        is a 4-tuple of integer Chimera indices.

    Examples
    --------
    >>> G = dnx.chimera_graph(1, 1, 4)
    >>> chimera_indices = dnx.find_chimera_indices(G)

    >>> G = nx.Graph()
    >>> G.add_edges_from([(0, 2), (1, 2), (1, 3), (0, 3)])
    >>> chimera_indices = dnx.find_chimera_indices(G)
    >>> nx.set_node_attributes(G, chimera_indices, 'chimera_index')

    """

    # if the nodes are orderable, we want the lowest-order one.
    try:
        nlist = sorted(G.nodes)
    except TypeError:
        nlist = G.nodes()

    n_nodes = len(nlist)

    # create the object that will store the indices
    chimera_indices = {}

    # ok, let's first check for the simple cases
    if n_nodes == 0:
        return chimera_indices
    elif n_nodes == 1:
        raise DWaveNetworkXException(
            'Singleton graphs are not Chimera-structured')
    elif n_nodes == 2:
        return {nlist[0]: (0, 0, 0, 0), nlist[1]: (0, 0, 1, 0)}

    # next, let's get the bicoloring of the graph; this raises an exception if the graph is
    # not bipartite
    coloring = color(G)

    # we want the color of the node to be the u term in the Chimera-index, so we want the
    # first node in nlist to be color 0
    if coloring[nlist[0]] == 1:
        coloring = {v: 1 - coloring[v] for v in coloring}

    # we also want the diameter of the graph
    # claim: diameter(G) == m + n for |G| > 2
    dia = diameter(G)

    # we have already handled the |G| <= 2 case, so we know, for diameter == 2, that the Chimera
    # graph is a single tile
    if dia == 2:
        shore_indices = [0, 0]

        for v in nlist:
            u = coloring[v]
            chimera_indices[v] = (0, 0, u, shore_indices[u])
            shore_indices[u] += 1

        return chimera_indices

    # NB: max degree == shore size <==> one tile

    raise Exception('not yet implemented for Chimera graphs with more than one tile')


class chimera_coordinates(object):
    """Provides coordinate converters for the chimera indexing scheme.

    Parameters
    ----------
    m : int
        The number of rows in the Chimera lattice.
    n : int, optional (default m)
        The number of columns in the Chimera lattice.
    t : int, optional (default 4)
        The size of the shore within each Chimera tile.

    Examples
    --------

    Convert between Chimera coordinates and linear indices directly

    >>> coords = dnx.chimera_coordinates(16, 16, 4)
    >>> coords.chimera_to_linear((0, 2, 0, 1))
    17
    >>> coords.linear_to_chimera(17)
    (0, 2, 0, 1)

    Construct a new graph with the coordinate labels

    >>> C16 = dnx.chimera_graph(16)
    >>> coords = dnx.chimera_coordinates(16)
    >>> G = nx.Graph()
    >>> G.add_nodes_from(coords.iter_linear_to_chimera(C16.nodes))
    >>> G.add_edges_from(coords.iter_linear_to_chimera_pairs(C16.edges))

    See also
    --------
    :func:`.chimera_graph` : Describes the various conventions.

    """
    def __init__(self, m, n=None, t=None):
        self.args = m, m if n is None else n, 4 if t is None else t

    def int(self, q):
        """Deprecated alias for `chimera_to_linear`."""
        msg = ('chimera_coordinates.int is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.chimera_to_linear instead')
        warnings.warn(msg, DeprecationWarning)
        return self.chimera_to_linear(q)

    def chimera_to_linear(self, q):
        """Convert a 4-term Chimera coordinate to a linear index.

        Parameters
        ----------
        q : 4-tuple
            Chimera coordinate.

        Examples
        --------
        >>> dnx.chimera_coordinates(16).chimera_to_linear((2, 2, 0, 0))
        272

        """
        i, j, u, k = q
        m, n, t = self.args
        return ((n*i + j)*2 + u)*t + k

    def tuple(self, r):
        """Deprecated alias for `linear_to_chimera`."""
        msg = ('chimera_coordinates.tuple is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.linear_to_chimera instead')
        warnings.warn(msg, DeprecationWarning)
        return self.linear_to_chimera(r)

    def linear_to_chimera(self, r):
        """Convert a linear index to a 4-term Chimera coordinate.

        Parameters
        ----------
        r : int
            Linear index.

        Examples
        --------
        >>> dnx.chimera_coordinates(16).linear_to_chimera(272)
        (2, 2, 0, 0)
        """
        m, n, t = self.args
        r, k = divmod(r, t)
        r, u = divmod(r, 2)
        i, j = divmod(r, n)
        return i, j, u, k

    def ints(self, qlist):
        """Deprecated alias for `iter_chimera_to_linear`."""
        msg = ('chimera_coordinates.ints is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.iter_chimera_to_linear instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_chimera_to_linear(qlist)

    def iter_chimera_to_linear(self, qlist):
        """Return an iterator converting a sequence of 4-term Chimera
        coordinates to linear indices.
        """
        m, n, t = self.args
        for (i, j, u, k) in qlist:
            yield ((n*i + j)*2 + u)*t + k

    def tuples(self, rlist):
        """Deprecated alias for `iter_linear_to_chimera`."""
        msg = ('chimera_coordinates.tuples is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.iter_linear_to_chimera instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_linear_to_chimera(rlist)

    def iter_linear_to_chimera(self, rlist):
        """Return an iterator converting a sequence of linear indices to 4-term
        Chimera coordinates.
        """
        m, n, t = self.args
        for r in rlist:
            r, k = divmod(r, t)
            r, u = divmod(r, 2)
            i, j = divmod(r, n)
            yield i, j, u, k

    @staticmethod
    def _pair_repack(f, plist):
        """Flattens a sequence of pairs to pass through `f`, and then
        re-pairs the result.
        """
        ulist = f(u for p in plist for u in p)
        for u in ulist:
            v = next(ulist)
            yield u, v

    def int_pairs(self, plist):
        """Deprecated alias for `iter_chimera_to_linear_pairs`."""
        msg = ('chimera_coordinates.int_pairs is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.iter_chimera_to_linear_pairs instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_chimera_to_linear_pairs(plist)

    def iter_chimera_to_linear_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of 4-term Chimera
        coordinates to pairs of linear indices.
        """
        return self._pair_repack(self.iter_chimera_to_linear, plist)

    def tuple_pairs(self, plist):
        """Deprecated alias for `iter_linear_to_chimera_pairs`."""
        msg = ('chimera_coordinates.tuple_pairs is deprecated and will be removed in '
               'dwave-networkx 0.9.0, please use '
               'chimera_coordinates.iter_linear_to_chimera_pairs instead')
        warnings.warn(msg, DeprecationWarning)
        return self.iter_linear_to_chimera_pairs(plist)

    def iter_linear_to_chimera_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of linear indices
        to pairs of 4-term Chimera coordinates.
        """
        return self._pair_repack(self.iter_linear_to_chimera, plist)

    def graph_to_linear(self, g):
        """Return a copy of the graph g relabeled to have linear indices"""
        labels = g.graph.get('labels')
        if labels == 'int':
            return g.copy()
        elif labels == 'coordinate':
            return chimera_graph(
                g.graph['rows'],
                n=g.graph['columns'],
                t=g.graph['tile'],
                node_list=self.iter_chimera_to_linear(g),
                edge_list=self.iter_chimera_to_linear_pairs(g.edges),
                data=g.graph['data'],
            )
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.chimera_graph."
            )

    def graph_to_chimera(self, g):
        """Return a copy of the graph g relabeled to have chimera coordinates"""
        labels = g.graph.get('labels')
        if labels == 'int':
            return chimera_graph(
                g.graph['rows'],
                n=g.graph['columns'],
                t=g.graph['tile'],
                node_list=self.iter_linear_to_chimera(g),
                edge_list=self.iter_linear_to_chimera_pairs(g.edges),
                data=g.graph['data'],
                coordinates=True,
            )
        elif labels == 'coordinate':
            return g.copy()
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.chimera_graph."
            )

class __chimera_coordinates_cache_dict(dict):
    """An internal-use cached factory for `chimera_coordinates` objects"""

    def __missing__(self, key):
        self[key] = val = chimera_coordinates(*key)
        return val


_chimera_coordinates_cache = __chimera_coordinates_cache_dict()

def linear_to_chimera(r, m, n=None, t=None):
    """Convert the linear index `r` into a chimera index.

    Parameters
    ----------
    r : int
        The linear index value.
    m : int
        Number of rows in the Chimera lattice.
    n : int (optional, default m)
        Number of columns in the Chimera lattice.
    t : int (optional, default 4)
        Size of the shore within each Chimera tile.


    Returns
    -------
    i : int
        The column of the Chimera index's unit cell associated with `r`.
    j : int
        The row of the Chimera index's unit cell associated with `r`.
    u : int
        Whether the index is even (0) or odd (1); the side of the bi-partite
        graph of the Chimera unit cell.
    k : int
        Index into the Chimera unit cell.

    Examples
    --------

    >>> G = dnx.linear_to_chimera(212, 8, 8, 4)
    (3, 2, 1, 0)

    """
    return _chimera_coordinates_cache[m, n, t].linear_to_chimera(r)


def chimera_to_linear(i, j, u, k, m, n=None, t=None):
    """Convert the chimera index `(i, j, u, k)` into a linear index.

    Parameters
    ----------
    i : int
        The column of the Chimera index's unit cell associated with `r`.
    j : int
        The row of the Chimera index's unit cell associated with `r`.
    u : int
        Whether the index is even (0) or odd (1); the side of the bi-partite
        graph of the Chimera unit cell.
    k : int
        Index into the Chimera unit cell.
    m : int
        Number of rows in the Chimera lattice.
    n : int (optional, default m)
        Number of columns in the Chimera lattice.
    t : int (optional, default 4)
        Size of the shore within each Chimera tile.


    Returns
    -------
    r : int
        The linear index node label corresponding to `(i, j, u, k)`.

    Examples
    --------

    >>> G = dnx.chimera_to_linear(3, 2, 1, 0, 8, 8, 4)
    212

    """
    return _chimera_coordinates_cache[m, n, t].chimera_to_linear((i, j, u, k))


def _chimera_sublattice_mapping(source_to_chimera, chimera_to_target, offset):
    """Constructs a mapping from one chimera graph to another, via an offset.
    This function is used by chimera_sublattice_mappings, and serves to 
    construct a closure that is stable under iteration therein.

    Parameters
    ----------
        source_to_chimera : function
            A function mapping a source node to a chimera-coordinate 
        chimera_to_target: function
            A function mapping a chimera coordinate to a target nodes
        offset : tuple (int, int)
            A pair of ints representing the y- and x-offset of the sublattice

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Chimera
            graph to the target Chimera graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.
        
    """
    y_offset, x_offset = offset

    def mapping(q):
        y, x, u, k = source_to_chimera(q)
        return chimera_to_target((y + y_offset, x + x_offset, u, k))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping


def chimera_sublattice_mappings(source, target, offset_list=None):
    """Yields mappings from a Chimera graph into a larger Chimera graph.
    
    A sublattice mapping is a function from nodes of a
    ``chimera_graph(m_s, n_s, t)`` to nodes of a ``chimera_graph(m_t, n_t, t)``
    with ``m_s <= m_t`` and ``n_s <= n_t``.  This is used to identify subgraphs 
    of the target Chimera graphs which are isomorphic to the source Chimera
    graph.  However, if the target graph is not of perfect yield, these 
    functions do not generally produce isomorphisms (for example, if a node is
    missing in the target graph, it may still appear in the image of the source
    graph).
    
    Note that we do not produce mappings between Chimera graphs of different
    tile parameters, and the mappings produced are not exhaustive.  The mappings
    take the form
    
        ``(y, x, u, k) -> (y+y_offset, x+x_offset, u, k)``
        
    preserving the orientation and tile index of nodes.  We use the notation of
    Chimera coordinates above, but either or both of the target graph may have
    integer or coordinate labels.
    
    Academic note: the full group of isomorphisms of a Chimera graph includes 
    mappings which permute tile indices on a per-row and per-column basis, in
    addition to reflections and rotations of the grid of unit cells where 
    rotations by 90 and 270 degrees induce a change in orientation.  The full
    set of sublattice mappings would take those isomorphisms into account; we do
    not undertake that complexity here.
    
    Parameters
    ----------
        source : NetworkX Graph
            The Chimera graph that nodes are input from
        target : NetworkX Graph
            The Chimera graph that nodes are input from
        offset_list : iterable (tuple), optional (default None)
            An iterable of offsets.  This can be used to reconstruct a set of
            mappings, as the offset used to generate a single mapping is stored
            in the ``offset`` attribute of that mapping.
            
    Yields
    ------
        mapping : function
            A function from nodes of the source graph, to nodes of the target
            graph.  The offset used to generate this mapping is stored in
            ``mapping.offset`` -- these can be collected and passed into 
            ``offset_list`` in a later session.

    """
    if not (source.graph.get('family') == target.graph.get('family') == 'chimera'):
        raise ValueError("source and target graphs must be Chimera graphs constructed by dwave_networkx.chimera_graph")
    
    t = source.graph['tile']
    if t != target.graph['tile']:
        raise ValueError("Cannot construct a sublattice mappings between Chimera graphs with different tile parameters")

    m_s = source.graph['rows']
    n_s = source.graph['columns']
    labels_s = source.graph['labels']
    if labels_s == 'int':
        source_to_chimera = _chimera_coordinates_cache[m_s, n_s, t].linear_to_chimera
    elif labels_s == 'coordinate':
        def source_to_chimera(q):
            return q
    else:
        raise ValueError(f"Chimera node labeling {labels_s} not recognized")

    m_t = target.graph['rows']
    n_t = target.graph['columns']
    labels_t = target.graph['labels']
    if labels_t == 'int':
        chimera_to_target = _chimera_coordinates_cache[m_t, n_t, t].chimera_to_linear
    elif labels_t == 'coordinate':
        def chimera_to_target(q):
            return q
    else:
        raise ValueError(f"Chimera node labeling {labels_t} not recognized")
    
    if offset_list is None:
        y_offsets = range(m_t - m_s + 1)
        x_offsets = range(n_t - n_s + 1)
        offset_list = product(y_offsets, x_offsets)

    for offset in offset_list:
        yield _chimera_sublattice_mapping(source_to_chimera, chimera_to_target, offset)




# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Generators for some graphs derived from the D-Wave System.
"""
import re
import warnings
from itertools import product

import networkx as nx

from dwave_networkx.exceptions import DWaveNetworkXException

from .chimera import _chimera_coordinates_cache

__all__ = ['zephyr_graph',
           'zephyr_coordinates',
           'zephyr_sublattice_mappings',
           ]


def zephyr_graph(m, t=4, create_using=None, node_list=None, edge_list=None,
                   data=True, coordinates=False):
    """
    Creates a Zephyr graph with grid parameter ``m`` and tile parameter ``t``.

    The Zephyr topology is described in [brk]_.

    Parameters
    ----------
    m : int
        Grid parameter for the Zephyr lattice.
    t : int
        Tile parameter for the Zephyr lattice.
    create_using : Graph, optional (default None)
        If provided, this graph is cleared of nodes and edges and filled
        with the new graph. Usually used to set the type of the graph.
    node_list : iterable, optional (default None)
        Iterable of nodes in the graph. If None, calculated from ``m``.
        Note that this list is used to remove nodes, so only specified nodes
        that belong to the base node set (described in the ``coordinates``
        parameter) are added.
    edge_list : iterable, optional (default None)
        Iterable of edges in the graph. If None, edges are generated as
        described below. The nodes in each edge must be labeled according to the
        ``coordinates`` parameter.
    data : bool, optional (default True)
        If True, adds to each node an attribute with a format that depends on
        the ``coordinates`` parameter: a 5-tuple ``'zephyr_index'`` if
        ``coordinates`` is False and an integer ``'linear_index'`` if ``coordinates``
        is True.
    coordinates : bool, optional (default False)
        If True, node labels are 5-tuple Zephyr indices.

    Returns
    -------
    G : NetworkX Graph
        A Zephyr lattice for grid parameter ``m`` and tile parameter ``t``.


    The maximum degree of this graph is :math:`4t+4`. The number of nodes is
    given by

        * ``zephyr_graph(m, t)``: :math:`4tm(2m+1)`

    The number of edges depends on parameter settings,

        * ``zephyr_graph(1, t)``: :math:`2t(8t+3)`
        * ``zephyr_graph(m, t)``: :math:`2t((8t+8)m^2-2m-3)`  if m > 1

    A Zephyr lattice is a graph minor of a lattice similar to Chimera, where
    unit tiles have odd couplers similar to Pegasus graphs. In its most
    general definition, prelattice :math:`Q(2m+1)` contains nodes of the form

        * vertical nodes: :math:`(i, j, 0, k)` with :math:`0 <= k < 2t`
        * horizontal nodes: :math:`(i, j, 1, k)` with :math:`0 <= k < 2t`

    for :math:`0 <= i < 2m+1` and :math:`0 <= j < 2m+1`, and edges of the form

        * external: :math:`(i, j, u, k)` ~ :math:`(i+u, j+1-u, u, k)`
        * internal: :math:`(i, j, 0, k)` ~ :math:`(i, j, 1, h)`
        * odd: :math:`(i, j, u, 2k)` ~ :math:`(i, j, u, 2k+1)`

    The minor---a Zephyr lattice---is constructed by contracting pairs of
    external edges::

        I(0, w, k, j, z) = [(2*z+j, w, 0, 2*k+j), (2*z+1+j, w, 0, 2*k+j)]
        I(1, w, k, j, z) = [(w, 2*z+j, 1, 2*k+j), (w, 2*z+1+j, 1, 2*k+j)]

    and deleting the prelattice nodes of any pair not fully contained in
    :math:`Q(2m+1)`.

    The *Zephyr index* of a node in a Zephyr lattice, :math:`(u, w, k, j, z)`,
    can be interpreted as:

        * :math:`u`: qubit orientation (0 = vertical, 1 = horizontal)
        * :math:`w`: orthogonal major offset; :math:`0 <= w < 2m+1`
        * :math:`k`: orthogonal secondary offset; :math:`0 <= k < t`
        * :math:`j`: orthogonal minor offset; :math:`0 <= j < 2`
        * :math:`z`: parallel offset; :math:`0 <= z < m`

    Edges in the minor have the form

        * external: :math:`(u, w, k, j, z)` ~ :math:`(u, w, k, j, z+1)`
        * odd: :math:`(u, w, 2k, z)` ~ :math:`(u, w, 2k+1, z-a)`
        * internal: :math:`(0, 2w+1-a, k, j, z-jb)` ~ :math:`(1, 2z+1-b, h, i, w-ia)`

    for :math:`0 <= a < 2` and :math:`0 <= b < 2`, where internal edges only exist when

        1. :math:`0 <= 2w+1-a < 2m+1`,
        2. :math:`0 <= 2z+1-a < 2m+1`,
        3. :math:`0 <= z-jb < m`, and
        4. :math:`0 <= w-ia < m`.

    Linear indices are computed from Zephyr indices by the formula::

        q = (((u * (2 * m + 1) + w) * t + k) * 2 + j) * m + z


    Examples
    --------
    >>> G = dnx.zephyr_graph(2)
    >>> G.nodes(data=True)[(0, 0, 0, 0, 0)]    # doctest: +SKIP
    {'linear_index': 0}

    References
    ----------
    .. [brk] Boothby, Raymond, King, Zephyr Topology of D-Wave Quantum
        Processors, October 2021.
        https://dwavesys.com/media/fawfas04/14-1056a-a_zephyr_topology_of_d-wave_quantum_processors.pdf
    """
    G = nx.empty_graph(0, create_using)
    m = int(m)
    t = int(t)

    G.name = "zephyr_graph(%s, %s)" % (m, t)

    M = 2*m+1

    if coordinates:
        def label(*q):
            return q
        labels = 'coordinate'
    else:
        labels = 'int'
        def label(u, w, k, j, z):
            return (((u * M + w) * t + k) * 2 + j) * m + z

    construction = (("family", "zephyr"), ("rows", m), ("columns", m),
                    ("tile", t), ("data", data), ("labels", labels))

    G.graph.update(construction)

    if edge_list is None:
        #external edges
        G.add_edges_from((label(u, w, k, j, z), label(u, w, k, j, z + 1))
                         for u, w, k, j, z in product(
                            (0, 1), range(M), range(t), (0, 1), range(m-1)
                         ))

        #odd edges
        G.add_edges_from((label(u, w, k, 0, z), label(u, w, k, 1, z-a))
                         for u, w, k, a in product(
                            (0, 1), range(M), range(t), (0, 1)
                         )
                         for z in range(a, m))

        #internal edges
        G.add_edges_from((label(0, 2*w+1+a*(2*i-1), k, j, z), label(1, 2*z+1+b*(2*j-1), h, i, w))
                         for w, z, h, k, i, j, a, b in product(
                            range(m), range(m), range(t), range(t), (0, 1), (0, 1), (0, 1), (0, 1)
                         ))

    else:
        G.add_edges_from(edge_list)

    if node_list is not None:
        nodes = set(node_list)
        G.remove_nodes_from(set(G) - nodes)
        G.add_nodes_from(nodes)  # for singleton nodes

    if data:
        if coordinates:
            def fill_data():
                d = get_node_data((u, w, k, j, z))
                if d is not None:
                    d['linear_index'] = v

        else:
            def fill_data():
                d = get_node_data(v)
                if d is not None:
                    d['zephyr_index'] = (u, w, k, j, z)

        v = 0
        get_node_data = G.nodes.get
        for u in range(2):
            for w in range(M):
                for k in range(t):
                    for j in (0, 1):
                        for z in range(m):
                            fill_data()
                            v += 1

    return G


# Developer note: we could implement a function that creates the iter_*_to_* and
# iter_*_to_*_pairs methods just-in-time, but there are a small enough number
# that for now it makes sense to do them by hand.
class zephyr_coordinates(object):
    """Provides coordinate converters for the Zephyr indexing schemes.

    Parameters
    ----------
    m : int
        Grid parameter for the Zephyr lattice.
    t : int
        Tile parameter for the Zephyr lattice; must be even.

    See also
    --------
    :func:`.zephyr_graph` : Describes the various coordinate conventions.

    """
    def __init__(self, m, t=4):
        self.args = m, 2 * m + 1, t

    def zephyr_to_linear(self, q):
        """Convert a 5-term Zephyr coordinate into a linear index.

        Parameters
        ----------
        q : 5-tuple
            Zephyr coordinate.

        Examples
        --------
        >>> dnx.zephyr_coordinates(2).zephyr_to_linear((0, 1, 2, 1, 0))
        26
        """
        u, w, k, j, z = q
        m, M, t = self.args
        return (((u * M + w) * t + k) * 2 + j) * m + z

    def linear_to_zephyr(self, r):
        """Convert a linear index into a 5-term Zephyr coordinate.

        Parameters
        ----------
        r : int
            Linear index.

        Examples
        --------
        >>> dnx.zephyr_coordinates(2).linear_to_zephyr(137)
        (1, 3, 2, 0, 1)

        """
        m, M, t = self.args
        r, z = divmod(r, m)
        r, j = divmod(r, 2)
        r, k = divmod(r, t)
        u, w = divmod(r, M)
        return u, w, k, j, z

    def iter_zephyr_to_linear(self, qlist):
        """Return an iterator converting a sequence of 5-term Zephyr
        coordinates to linear indices.
        """
        m, M, t = self.args
        for (u, w, k, j, z) in qlist:
            yield (((u * M + w) * t + k) * 2 + j) * m + z

    def iter_linear_to_zephyr(self, rlist):
        """Return an iterator converting a sequence of linear indices to 5-term
        Zephyr coordinates.
        """
        m, M, t = self.args
        for r in rlist:
            r, z = divmod(r, m)
            r, j = divmod(r, 2)
            r, k = divmod(r, t)
            u, w = divmod(r, M)
            yield u, w, k, j, z

    @staticmethod
    def _pair_repack(f, plist):
        """Flattens a sequence of pairs to pass through `f`, and then
        re-pairs the result.
        """
        ulist = f(u for p in plist for u in p)
        for u in ulist:
            v = next(ulist)
            yield u, v

    def iter_zephyr_to_linear_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of 5-term Zephyr
        coordinates to pairs of linear indices.
        """
        return self._pair_repack(self.iter_zephyr_to_linear, plist)

    def iter_linear_to_zephyr_pairs(self, plist):
        """Return an iterator converting a sequence of pairs of linear indices
        to pairs of 5-term Zephyr coordinates.
        """
        return self._pair_repack(self.iter_linear_to_zephyr, plist)

    def graph_to_linear(self, g):
        """Return a copy of the graph g relabeled to have linear indices"""
        labels = g.graph.get('labels')
        if labels == 'int':
            return g.copy()
        elif labels == 'coordinate':
            nodes = self.iter_zephyr_to_linear(g)
            edges = self.iter_zephyr_to_linear_pairs(g.edges)
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.zephyr_graph."
            )

        return zephyr_graph(
            g.graph['rows'],
            t = g.graph['tile'],
            node_list=nodes,
            edge_list=edges,
            data=g.graph['data'],
        )

    def graph_to_zephyr(self, g):
        """Return a copy of the graph g relabeled to have zephyr coordinates"""
        labels = g.graph.get('labels')
        if labels == 'int':
            nodes = self.iter_linear_to_zephyr(g)
            edges = self.iter_linear_to_zephyr_pairs(g.edges)
        elif labels == 'coordinate':
            return g.copy()
        else:
            raise ValueError(
                f"Node labeling {labels} not recognized.  Input must be generated by dwave_networkx.zephyr_graph."
            )

        return zephyr_graph(
            g.graph['rows'],
            t=g.graph['tile'],
            node_list=nodes,
            edge_list=edges,
            data=g.graph['data'],
            coordinates=True,
        )


class __zephyr_coordinates_cache_dict(dict):
    """An internal-use cached factory for `zephyr_coordinates` objects"""

    def __missing__(self, key):
        self[key] = val = zephyr_coordinates(*key)
        return val


_zephyr_coordinates_cache = __zephyr_coordinates_cache_dict()


def _zephyr_zephyr_sublattice_mapping(source_to_zephyr, zephyr_to_target, offset):
    """Constructs a mapping from a Zephyr graph to a Zephyr graph, via an offset.
    This function is used by zephyr_sublattice_mappings, and serves to construct
    a closure that is stable under iteration therein.

    The mappings implemented by this function interpret offsets in the grid of
    the Chimera(2m+1, 2m+1, 2*t) graphs underlying the source and tartget Zephyr
    graphs.  The formulas (see implementation) are somewhat complex, because

        * a shift by a y-unit induces a reversal of the orthogonal minor offset
            (j index) of vertical qubits,
        * a shift by an x-unit induces a reversal of the j index of horizontal
            qubits, and
        * a shift by a unit parallel to a qubit is equivalent to 1/2-unit shift
            in the z-direction (but z-coordinates are integral) which is
            mediated by the j index.

    Parameters
    ----------
        source_to_zephyr : function
            A function mapping a source node to a zephyr coordinate
        zephyr_to_target: function
            A function mapping a zephyr coordinate to a target node
        offset : tuple (int, int)
            A pair of ints representing the y- and x-offset of the sublattice

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Zephyr
            graph to the target Zephyr graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.

    """
    y_offset, x_offset = offset

    delta = [
        [y_offset % 2, x_offset, y_offset],
        [x_offset % 2, y_offset, x_offset],
    ]

    def mapping(q):
        u, w, k, j, z = source_to_zephyr(q)
        dj, dw, dz = delta[u]
        return zephyr_to_target((u, w + dw, k, j ^ dj, z + (dz + j) // 2))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping

def _single_chimera_zephyr_sublattice_mapping(source_to_chimera, zephyr_to_target, offset):
    """Constructs a mapping from a Chimera graph to a Zephyr graph, via an offset.
    This function is used by zephyr_sublattice_mappings, and serves to construct
    a closure that is stable under iteration therein.

    The mappings implemented by this function view a ``chimera(2*m, 2*m, t)`` as
    a subgraph of ``zephyr_graph(m, t)`` through the mapping

        (2*y+j, x, 0, k) -> (0, x, k, j, y)
        (y, 2*x+j, 1, k) -> (1, y, k, j, x)

    which interprets odd couplers of Zephyr as external couplers of Chimera.
    The above is a slight simplification of matters; it is the simplest of a
    family of :math:`(t+1)^2` offsets (see how ``k_offset0`` and ``k_offset``
    are used in the implementation).

    Additionally, the sublattice represented  by the source graph can have x-
    and y-offsets into the chimera graph above, as with ordinary Chimera
    subgraph mappings.

    Parameters
    ----------
        source_to_chimera : function
            A function mapping a source node to a chimera coordinate
        zephyr_to_target: function
            A function mapping a zephyr coordinate to a target node
        offset : tuple (int, int, int, int, int)
            A tuple of ints (t, k_offset0, k_offset1, y_offset, x_offset)
            defining the sublattice mapping

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Zephyr
            graph to the target Zephyr graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.

    """
    t, y_offset, x_offset, k_offset0, k_offset1 = offset

    def mapping(q):
        y, x, u, k = source_to_chimera(q)
        if u:
            dw, k = divmod(k + k_offset1, t)
            z, j = divmod(x + x_offset, 2)
            return zephyr_to_target((u, y + y_offset + dw, k, j, z))
        else:
            dw, k = divmod(k + k_offset0, t)
            z, j = divmod(y + y_offset, 2)
            return zephyr_to_target((u, x + x_offset + dw, k, j, z))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping

def _double_chimera_zephyr_sublattice_mapping(source_to_chimera, zephyr_to_target, offset):
    """Constructs a mapping from a Chimera graph to a Zephyr graph, via an offset.
    This function is used by zephyr_sublattice_mappings, and serves to construct
    a closure that is stable under iteration therein.

    The mappings implemented by this function view a ``chimera(m, m, 2*t)`` as
    a subgraph of ``zephyr_graph(m, t)`` through the mappings

        (y, x, 0, k) -> (0, x, k, j0, y)
        (y, x, 1, k) -> (1, y, k, j1, x)

    where j0 and j1 are each 0 or 1.  Additionally, the sublattice represented
    by the source graph can have x- and y-offsets into the chimera graph above,
    as with ordinary Chimera subgraph mappings.

    Parameters
    ----------
        source_to_chimera : function
            A function mapping a source node to a chimera coordinate
        zephyr_to_target: function
            A function mapping a zephyr coordinate to a target node
        offset : tuple (int, int, int, int, int)
            A tuple of ints (t, j0, j1, y_offset, x_offset) defining the
            sublattice mapping

    Returns
    -------
        mapping : function
            The function implementing the mapping from the source Zephyr
            graph to the target Zephyr graph.  We store ``offset`` in the
            attribute ``mapping.offset`` for later reconstruction.

    """
    t, y_offset, x_offset, j0, j1 = offset
    def mapping(q):
        y, x, u, k = source_to_chimera(q)
        wz, kz = divmod(k, t)
        if u:
            return zephyr_to_target((u, 2 * (y + y_offset) + j0 + wz, kz, j1, x + x_offset))
        else:
            return zephyr_to_target((u, 2 * (x + x_offset) + j1 + wz, kz, j0, y + y_offset))

    #store the offset in the mapping, so the user can reconstruct it
    mapping.offset = offset

    return mapping


def zephyr_sublattice_mappings(source, target, offset_list=None):
    """Yields mappings from a Chimera or Zephyr graph into a Zephyr graph.

    A sublattice mapping is a function from nodes of

        * a ``zephyr_graph(m_s, t)`` to nodes of a ``zephyr_graph(m_t, t)``
          where ``m_s <= m_t``,
        * a ``chimera_graph(m_s, n_s, t)`` to nodes of a ``zephyr_graph(m_t, t)``
          where ``m_s <= 2*m_t`` and ``n_s <= 2*m_t``, or
        * a ``chimera_graph(m_s, n_s, 2*t)`` to nodes of a ``zephyr_graph(m_t, t)``
          where ``m_s <= m_t`` and ``n_s <= m_t``, or

    This is used to identify subgraphs of the target Zephyr graphs which are
    isomorphic to the source graph. However, if the target graph is not of
    perfect yield, these functions do not generally produce isomorphisms (for
    example, if a node is missing in the target graph, it may still appear in
    the image of the source graph).

    Note that the tile parameter of Chimera graphs must be either the
    same or double that of the target Zephyr graphs; if both graphs are
    Zephyr graphs, the tile parameters must be the same. The mappings
    produced preserve the linear ordering of tile indices; see the
    ``_zephyr_zephyr_sublattice_mapping``,
    ``_double_chimera_zephyr_sublattice_mapping``, and
    ``_single_chimera_zephyr_sublattice_mapping`` internal functions for more
    details.

    Academic note: the full group of isomorphisms of a Chimera graph includes
    mappings which permute tile indices on a per-row and per-column basis, in
    addition to reflections and rotations of the grid of unit cells where
    rotations by 90 and 270 degrees induce a change in orientation.  The
    isomorphisms of Zephyr graphs permit permutations of major tile indices on a
    per-row and per-column basis, in addition to reflections of the grid which
    induce inversion of orthogonal minor offsets, and rotations which induce
    inversions of minor offsets and/or orientation. The full set of sublattice
    mappings would take those isomorphisms into account; we do not undertake
    that complexity here.

    Parameters
    ----------
        source : NetworkX Graph
            The Chimera or Zephyr graph that nodes are input from.
        target : NetworkX Graph
            The Zephyr graph that nodes are output to.
        offset_list : iterable (tuple), optional (default None)
            An iterable of offsets. This can be used to reconstruct a set of
            mappings, as the offset used to generate a single mapping is stored
            in the ``offset`` attribute of that mapping.

    Yields
    ------
        mapping : function
            A function from nodes of the source graph, to nodes of the target
            graph.  The offset used to generate this mapping is stored in
            ``mapping.offset`` -- these can be collected and passed into
            ``offset_list`` in a later session.

    """
    if target.graph.get('family') != 'zephyr':
        raise ValueError("source graphs must a Zephyr graph constructed by dwave_networkx.zephyr_graph")

    m_t = target.graph['rows']
    t = target.graph['tile']
    labels_t = target.graph['labels']
    if labels_t == 'int':
        zephyr_to_target = _zephyr_coordinates_cache[m_t, t].zephyr_to_linear
    elif labels_t == 'coordinate':
        def zephyr_to_target(q):
            return q
    else:
        raise ValueError(f"Zephyr node labeling {labels_t} not recognized")

    labels_s = source.graph['labels']
    if source.graph.get('family') == 'chimera':
        t_t = source.graph['tile']
        m_s = source.graph['rows']
        n_s = source.graph['columns']

        if t_t == t:
            make_mapping = _single_chimera_zephyr_sublattice_mapping
            if offset_list is None:
                krange = range(t+1)
                mrange = range(2*m_t - m_s + 1)
                nrange = range(2*m_t - n_s + 1)
                offset_list = product([t], mrange, nrange, krange, krange)
        elif t_t == 2*t:
            make_mapping = _double_chimera_zephyr_sublattice_mapping
            if offset_list is None:
                jrange = range(2)
                mrange = range(m_t - m_s + 1)
                nrange = range(m_t - n_s + 1)
                offset_list = product([t], mrange, nrange, jrange, jrange)
        else:
            raise ValueError(f"Cannot construct sublattice mappings from Chimera to this Zephyr graph unless the tile parameter of the chimera graph is {t} or {2*t}.")

        if labels_s == 'coordinate':
            def source_to_inner(q):
                return q
        elif labels_s == 'int':
            source_to_inner = _chimera_coordinates_cache[m_s, n_s, t_t].linear_to_chimera
        else:
            raise ValueError(f"Chimera node labeling {labels_s} not recognized")

    elif source.graph.get('family') == 'zephyr':
        m_s = source.graph['rows']
        if offset_list is None:
            mrange = range((2*m_t+1) - (2*m_s+1) + 1)
            offset_list = product(mrange, mrange)

        labels_s = source.graph['labels']
        if labels_s == 'int':
            source_to_inner = _zephyr_coordinates_cache[m_s, t].linear_to_zephyr
        elif labels_s == 'coordinate':
            def source_to_inner(q):
                return q
        else:
            raise ValueError(f"Zephyr node labeling {labels_s} not recognized")

        make_mapping = _zephyr_zephyr_sublattice_mapping

    else:
        raise ValueError("source graph must be a Chimera graph or Zephyr graph constructed by dwave_networkx.chimera_graph or dwave_networkx.zephyr_graph respectively")

    for offset in offset_list:
        yield make_mapping(source_to_inner, zephyr_to_target, offset)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx
import dimod
import dwave_networkx as dnx


class TestIsIndependentSet(unittest.TestCase):
    def test_empty(self):
        G = nx.Graph()

        self.assertTrue(dnx.is_independent_set(G, []))

    def test_K1(self):
        G = nx.complete_graph(1)

        self.assertTrue(dnx.is_independent_set(G, [0]))
        self.assertTrue(dnx.is_independent_set(G, []))

    def test_K2(self):
        G = nx.complete_graph(2)

        self.assertTrue(dnx.is_independent_set(G, [0]))
        self.assertTrue(dnx.is_independent_set(G, []))
        self.assertFalse(dnx.is_independent_set(G, [0, 1]))

    def test_path3(self):
        G = nx.path_graph(3)

        self.assertTrue(dnx.is_independent_set(G, [0]))
        self.assertTrue(dnx.is_independent_set(G, [0, 2]))
        self.assertTrue(dnx.is_independent_set(G, []))
        self.assertFalse(dnx.is_independent_set(G, [0, 1]))


class TestWeightedMaximumIndependentSet(unittest.TestCase):
    def test_empty(self):
        G = nx.Graph()

        Q = dnx.maximum_weighted_independent_set_qubo(G)

        self.assertEqual(Q, {})

    def test_K1_no_weights(self):
        G = nx.complete_graph(1)

        Q = dnx.maximum_weighted_independent_set_qubo(G)

        self.assertEqual(Q, {(0, 0): -1})

    def test_K1_weighted(self):
        G = nx.Graph()
        G.add_node(0, weight=.5)

        Q = dnx.maximum_weighted_independent_set_qubo(G)

        self.assertEqual(Q, {(0, 0): -1.})  # should be scaled to 1

    def test_K2_weighted(self):
        G = nx.Graph()
        G.add_node(0, weight=.5)
        G.add_node(1, weight=1)
        G.add_edge(0, 1)

        Q = dnx.maximum_weighted_independent_set_qubo(G, weight='weight')

        self.assertEqual(Q, {(0, 0): -.5, (1, 1): -1, (0, 1): 2.0})

    def test_K2_partially_weighted(self):
        G = nx.Graph()
        G.add_node(0, weight=.5)
        G.add_node(1)
        G.add_edge(0, 1)

        Q = dnx.maximum_weighted_independent_set_qubo(G, weight='weight')

        self.assertEqual(Q, {(0, 0): -.5, (1, 1): -1, (0, 1): 2.0})

    def test_path3_weighted(self):
        G = nx.path_graph(3)
        G.nodes[1]['weight'] = 2.1

        Q = dnx.maximum_weighted_independent_set_qubo(G, weight='weight')

        self.assertLess(dimod.qubo_energy({0: 0, 1: 1, 2: 0}, Q),
                        dimod.qubo_energy({0: 1, 1: 0, 2: 1}, Q))


class TestIndepSet(unittest.TestCase):

    def test_maximum_independent_set_basic(self):
        """Runs the function on some small and simple graphs, just to make
        sure it works in basic functionality.
        """
        G = dnx.chimera_graph(1, 2, 2)
        indep_set = dnx.maximum_independent_set(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_independent_set(G, indep_set))

        G = nx.path_graph(5)
        indep_set = dnx.maximum_independent_set(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_independent_set(G, indep_set))

    def test_maximum_independent_set_weighted(self):
        weight = 'weight'
        G = nx.path_graph(6)

        # favor odd nodes
        nx.set_node_attributes(G, {node: node % 2 + 1 for node in G}, weight)
        indep_set = dnx.maximum_weighted_independent_set(G, weight, dimod.ExactSolver())
        self.assertEqual(set(indep_set), {1, 3, 5})

        # favor even nodes
        nx.set_node_attributes(G, {node: (node + 1) % 2 + 1 for node in G}, weight)
        indep_set = dnx.maximum_weighted_independent_set(G, weight, dimod.ExactSolver())
        self.assertEqual(set(indep_set), {0, 2, 4})

        # make nodes 1 and 4 likely
        nx.set_node_attributes(G, {0: 1, 1: 3, 2: 1, 3: 1, 4: 3, 5: 1}, weight)
        indep_set = dnx.maximum_weighted_independent_set(G, weight, dimod.ExactSolver())
        self.assertEqual(set(indep_set), {1, 4})

    def test_default_sampler(self):
        G = nx.complete_graph(5)

        dnx.set_default_sampler(dimod.ExactSolver())
        self.assertIsNot(dnx.get_default_sampler(), None)
        indep_set = dnx.maximum_independent_set(G)
        dnx.unset_default_sampler()
        self.assertEqual(dnx.get_default_sampler(), None, "sampler did not unset correctly")

    def test_dimod_vs_list(self):
        G = nx.path_graph(5)

        indep_set = dnx.maximum_independent_set(G, dimod.ExactSolver())
        indep_set = dnx.maximum_independent_set(G, dimod.SimulatedAnnealingSampler())



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import itertools

import networkx as nx

import dimod

import dwave_networkx as dnx


class TestMinVertexColor(unittest.TestCase):

    def test_5path(self):
        G = nx.path_graph(5)
        coloring = dnx.min_vertex_coloring(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))
        self.assertEqual(len(set(coloring.values())), 2)  # bipartite

    def test_odd_cycle_graph(self):
        """Graph that is an odd circle"""
        G = nx.cycle_graph(5)
        coloring = dnx.min_vertex_coloring(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))

    def test_disconnected_graph(self):
        """One edge and one disconnected node"""
        G = nx.path_graph(2)
        G.add_node(3)

        coloring = dnx.min_vertex_coloring(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))

    def test_disconnected_cycle_graph(self):
        G = nx.complete_graph(3)  # odd 3-cycle
        G.add_node(4)  # floating node
        coloring = dnx.min_vertex_coloring(G, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))


class TestMinVertexColorQubo(unittest.TestCase):
    def test_chromatic_number(self):
        G = nx.cycle_graph('abcd')

        # when the chromatic number is fixed this is exactly vertex_color
        self.assertEqual(dnx.min_vertex_color_qubo(G, chromatic_lb=2,
                                                   chromatic_ub=2),
                         dnx.vertex_color_qubo(G, 2))


class TestVertexColor(unittest.TestCase):
    def test_4cycle(self):
        G = nx.cycle_graph('abcd')

        coloring = dnx.vertex_color(G, 2, dimod.ExactSolver())

        self.assertTrue(dnx.is_vertex_coloring(G, coloring))

    def test_4cycle_with_chord(self):
        G = nx.cycle_graph(4)
        G.add_edge(0, 2)

        # need 3 colors in this case
        coloring = dnx.vertex_color(G, 3, dimod.ExactSolver())

        self.assertTrue(dnx.is_vertex_coloring(G, coloring))

    def test_5cycle(self):
        G = nx.cycle_graph(5)
        coloring = dnx.vertex_color(G, 3, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))

    def test_disconnected_cycle_graph(self):
        G = nx.complete_graph(3)  # odd 3-cycle
        G.add_node(4)  # floating node
        coloring = dnx.vertex_color(G, 3, dimod.ExactSolver())
        self.assertTrue(dnx.is_vertex_coloring(G, coloring))


class TestVertexColorQUBO(unittest.TestCase):
    def test_single_node(self):
        G = nx.Graph()
        G.add_node('a')

        # a single color
        Q = dnx.vertex_color_qubo(G, ['red'])

        self.assertEqual(Q, {(('a', 'red'), ('a', 'red')): -1})

    def test_4cycle(self):
        G = nx.cycle_graph('abcd')

        Q = dnx.vertex_color_qubo(G, 2)

        sampleset = dimod.ExactSolver().sample_qubo(Q)

        # check that the ground state is a valid coloring
        ground_energy = sampleset.first.energy

        colorings = []
        for sample, en in sampleset.data(['sample', 'energy']):
            if en > ground_energy:
                break

            coloring = {}
            for (v, c), val in sample.items():
                if val:
                    coloring[v] = c

            self.assertTrue(dnx.is_vertex_coloring(G, coloring))

            colorings.append(coloring)

        # there are two valid colorings
        self.assertEqual(len(colorings), 2)

        self.assertEqual(ground_energy, -len(G))

    def test_num_variables(self):
        G = nx.Graph()
        G.add_nodes_from(range(15))

        Q = dnx.vertex_color_qubo(G, 7)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)
        self.assertEqual(len(bqm.quadratic), len(G)*7*(7-1)/2)

        # add one edge
        G.add_edge(0, 1)
        Q = dnx.vertex_color_qubo(G, 7)
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)
        self.assertEqual(len(bqm.quadratic), len(G)*7*(7-1)/2 + 7)

    def test_docstring_stats(self):
        # get a complex-ish graph
        G = nx.karate_club_graph()

        colors = range(10)

        Q = dnx.vertex_color_qubo(G, colors)

        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        self.assertEqual(len(bqm), len(G)*len(colors))
        self.assertEqual(len(bqm.quadratic), len(G)*len(colors)*(len(colors)-1)/2
                         + len(G.edges)*len(colors))
